/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of IDLBrowser.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#import "IDLItem.h"
#import "IDLFunction.h"
#import "IDLInterface.h"
#import "IDLStruct.h"
#import "IDLArray.h"
#import "IDLSequence.h"
#import "IDLUnion.h"
#import "IDLSwitchCase.h"
#import "IDLValueType_Private.h"
#import "IDLTCKind.h"
#import "CORBAObject.h"

@interface NSObject(ItemController)
- (void) invalidateItem:item;
- (void) setScope;
- (void) expandItem:(IDLItem*)item;
- (void) collapseItem:(IDLItem*)item;
@end
@interface NSObject(ValueWrapper)
- wrappedValue;
@end

@implementation IDLItem

NSString* IDLItemPboardType = @"IDLItemPboardType";
NSString* IDLItemLinkPboardType = @"IDLItemLinkPboardType";
NSColor* _classColor = nil;

+ (void) initialize
{
	if([IDLItem class] == self){
		_classColor = [[NSColor colorWithCalibratedRed:0.25 green:0.25 blue:0.5 alpha:1.0]retain];
//		[self setKeys:[NSArray arrayWithObjects:@"editable", nil]
//				triggerChangeNotificationsForDependentKey:@"itemEditable"];
	}
}

+ (NSSet*) keyPathsForValuesAffectingItemEditable
{
	return [NSSet setWithObject:@"editable"];
}

+ (NSColor*) classColor
{
	return _classColor;
}

+ (BOOL)automaticallyNotifiesObserversForKey:(NSString *)key
{
	return NO;
}

- initWithIDLType:(IDLDefinition*)def
{
	self = [super init];
	_type = [def retain];
	_value = [[def defaultItemValue]retain];
	[self startObservingValueChanges];
	return self;
}

- (void) dealloc
{
	[self stopObservingValueChanges];
	[_type release];
	[_value release];
	[super dealloc];
}

+ itemWithIDLType:(IDLDefinition*)def
{
	Class class = [def itemClass];
	return [[[class alloc]initWithIDLType:def]autorelease];
}

- (void) setItemName:(NSString*)name
{
	if(name != _name){
		[self willChangeValueForKey:@"itemName"];
		[_name release];
		_name = [name retain];
		[self didChangeValueForKey:@"itemName"];
	}
}

- (NSString*)itemName
{
	return _name;
}

- itemValue
{
	return _value;
}
		
- (void) _updateItems
{
	[_controller invalidateItem:self];
}

- (void) setItemValue:value
{
	[self willChangeValueForKey:@"itemValue"];
	[self setValue:value];
	[self didChangeValueForKey:@"itemValue"];
}

// the key "value" returns the item's real value
// while the key "itemValue" return s a value to be displayed
- value
{
	return _value;
}

// observers sign for the key "itemValue"
// using setValue: does the same but without notifiyng observers
- (void) setValue:value
{
	if(value != _value){
		[self stopObservingValueChanges];
		[_value release];
		_value = [value retain];
		[self startObservingValueChanges];
		[self _updateItems];
	}
}

- (void) startObservingValueChanges
{
}
- (void) stopObservingValueChanges
{
}

- (IDLDefinition*)itemType
{
	return _type;
}

- (unsigned) itemCount
{
	return 0;
}

- (NSArray*) items
{
	return nil;
}

- (BOOL) isLeaf
{
	return [_type isLeaf];
}

- (NSCell*) dataCell
{
	NSCell* cell = [_type itemDataCell];
	[cell setEditable:[self isEditable]];
	return cell;
}

- (void)resetCell:(NSTextFieldCell*)cell
{
	[_type resetItemCell:cell];
}

- (IDLInterface*)interfaceObject
{
	if([_type respondsToSelector:@selector(interfaceObject)])
		return [(IDLFunction*)_type interfaceObject];
	else
		return nil;
}

- (BOOL) isItemEditable
{
	return _flags.editable;
}

- (BOOL) isEditable
{
	return _flags.editable;
}

- (void) setEditable:(BOOL)flag
{
	[self willChangeValueForKey:@"editable"];
	_flags.editable = flag?YES:NO;
	[self didChangeValueForKey:@"editable"];
}

- (BOOL) shouldSetIDLType:(IDLDefinition*)type
{
	return NO;
}

- (BOOL) shouldInsertItemWithIDLType:(IDLDefinition*)type
{
	return NO;
}

- (void) setIDLType:(IDLDefinition*)def
{
	[self doesNotRecognizeSelector:_cmd];
}

- insertNewItemWithIDLType:(IDLDefinition*)def atIndex:(int)index
{
	[self doesNotRecognizeSelector:_cmd];
	return nil;
}

- (BOOL) shouldTakeValueFromItem:(IDLItem*)item
{
	return [self isEditable] && [[item itemType]isDerivedFromIDLType:_type];
}

- (BOOL) shouldInsertItem:(IDLItem*)item
{
	return NO;
}
- (void) takeValueFromItem:(IDLItem*)item
{
	if([[item itemType]isDerivedFromIDLType:_type])
		[self setValue:[item value] forKey:@"itemValue"];
}

- (void) insertItem:(IDLItem*)item atIndex:(int)index
{
	[self doesNotRecognizeSelector:_cmd];
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	[coder encodeObject:_type forKey:@"type"];
	[coder encodeObject:_value forKey:@"value"];
	[coder encodeObject:_name forKey:@"name"];
}

- initWithCoder:(NSCoder*)coder
{
	self = [super init];
	_type = [[coder decodeObjectForKey:@"type"]retain];
	_value = [[coder decodeObjectForKey:@"value"]retain];
	_name = [[coder decodeObjectForKey:@"name"]retain];
	// Do not start observing. This is responsibility of a parent after inserting this item
	return self;
}

- controller
{
	return _controller;
}

- (void)setController:controller
{
	_controller = controller;
}

- (NSString*) valueDescription
{
	return nil;
}

- (BOOL) isInTree
{
	return !_flags.notInTree;
}

- (void) setIsInTree:(BOOL)flag
{
	_flags.notInTree = flag?NO:YES;
}

- (BOOL) writeToPasteboard:(NSPasteboard*)pboard
{
	NSData* data = [NSKeyedArchiver archivedDataWithRootObject:self];
	[pboard declareTypes:[NSArray arrayWithObjects:IDLItemPboardType, IDLItemLinkPboardType, nil] owner:nil];
	[pboard setData:data forType:IDLItemPboardType];
	data = [NSData dataWithBytes:&self length:sizeof(id)];
	[pboard setData:data forType:IDLItemLinkPboardType];
	return YES;
}

@end

@implementation IDLAttributeItem

+ (BOOL)automaticallyNotifiesObserversForKey:(NSString *)key
{
	return NO;
}

- initWithIDLType:(IDLDefinition*)type
{
	self = [super init];
	_type = [type retain];
	_valueItem = [(IDLAttribute*)type createValueItemWithName:@"value"];
//	[_valueItem addOserver:self forKeyPath:@"itemValue" options:0 context:NULL];
	return self;
}

+ itemWithIDLType:(IDLDefinition*)type
{
	return [[[self alloc]initWithIDLType:type]autorelease];
}

- (void) dealloc
{
//	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem release];
	[_type release];
	[_name release];
	[super dealloc];
}

- (IDLInterface*)interfaceObject
{
	return [(IDLAttribute*)_type interfaceObject];
}

- (void) setItemName:(NSString*)name
{
	if(name != _name){
		[self willChangeValueForKey:@"itemName"];
		[_name release];
		_name = [name retain];
		[self didChangeValueForKey:@"itemName"];
	}
}

- (IDLDefinition*)itemType
{
	return _type;
}

- (NSString*)itemName
{
	return _name;
}

- (void) setItemValue:value
{
	[self willChangeValueForKey:@"itemValue"];
	[self setValue:value];
	[self didChangeValueForKey:@"itemValue"];
}

- (BOOL) isLeaf
{
	return [_type isLeaf];
}

- (void) setValue:value
{
	[_valueItem setValue:value];
}

- value
{
	return [_valueItem value];
}

- itemValue
{
	return [_valueItem itemValue];
}

- insertNewItemAtIndex:(int)index
{
	if([_valueItem respondsToSelector:_cmd])
		return [(IDLSequenceItem*)_valueItem insertNewItemAtIndex:index];
	else
		return nil;
}

- (void) removeItemAtIndex:(int)index
{
	if([_valueItem respondsToSelector:_cmd])
		[(IDLSequenceItem*)_valueItem removeItemAtIndex:index];
}

- (BOOL) shouldRemoveItem
{
	if([_valueItem respondsToSelector:_cmd])
		return [(IDLSequenceItem*)_valueItem shouldRemoveItem];
	else
		return NO;
}
	
- (BOOL) shouldInsertNewItem
{
	if([_valueItem respondsToSelector:_cmd])
		return [(IDLSequenceItem*)_valueItem shouldInsertNewItem];
	else
		return NO;
}

- (NSMethodSignature*) methodSignatureForSelector:(SEL)cmd
{
	return [_valueItem methodSignatureForSelector:cmd];
}
- (void) forwardInvocation:(NSInvocation*)inv
{
	[inv invokeWithTarget:_valueItem];
}

@end

@implementation IDLStructItem

- (BOOL) isItemEditable
{
	return NO;
}

- (unsigned) itemCount
{
	return [[(IDLStruct*)_type members]count];
}

- (void) startObservingValueChanges
{
	if(_value && !_flags.observing){
		[_type addValueObserver:self forValue:_value];
		_flags.observing = YES;
	}
}

- (void) stopObservingValueChanges
{
	if(_flags.observing){
		[_type removeValueObserver:self forValue:_value];
		_flags.observing = NO;
	}
}
		
- (NSArray*) items
{
	if(!_flags.itemsLoaded){
		_items = [[(IDLStruct*)_type itemArrayForItem:self]retain];
		if([_items count] > 0){
			NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
			[_items addObserver:self toObjectsAtIndexes:set forKeyPath:@"itemValue" 
				options:NSKeyValueObservingOptionNew|NSKeyValueObservingOptionOld context:NULL];
			[_items makeObjectsPerformSelector:@selector(setController:) withObject:[self controller]];
		}
		_flags.itemsLoaded = YES;
	}
	return _items;
}

- (IDLItem*) findItemWithItemName:(NSString*)name
{
	if(_flags.itemsLoaded){
		NSEnumerator* en = [_items objectEnumerator];
		IDLItem* item;
		while(item = [en nextObject])
			if([[item itemName]isEqualToString:name])
				break;
		return item;
	}else
		return nil;
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)object change:(NSDictionary*)change context:(void*)context
{
	if(object == _value){
		IDLItem* item = [self findItemWithItemName:keyPath];
		if(item){
			[item setValue:[_value valueForKeyPath:keyPath]];
			[_controller endEditing];
			[_controller invalidateItem:item];
		}
	}else
		[_value setValue:[object value] forKey:[object itemName]];
}

- itemValue
{
	return [_type name];
}

- (void) _updateItems
{
	NSEnumerator* en = [_items objectEnumerator];
	IDLItem* item;
	while(item = [en nextObject])
		[item setValue:[_value valueForKey:[item itemName]]];
}

- (NSString*) valueDescription
{
	return [_value description];
}

- (void) dealloc
{
	[self stopObservingValueChanges];
	NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
	[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
	[_items release];
	[super dealloc];
}

@end

@implementation IDLValueClassItem
- (unsigned) itemCount
{
	return [(IDLValueType*)_type memberCount];
}

- itemValue
{
	return [NSString stringWithFormat:@"%@ 0x%x", [_type name], _value];
}

@end

@implementation IDLArrayItem

- (void) startObservingValueChanges
{
	if(!_flags.observing){
		[_value addObserver:self forKeyPath:@"value" options:0 context:NULL];
		_flags.observing = YES;
	}
}
- (void) stopObservingValueChanges
{
	if(_flags.observing){
		[_value removeObserver:self forKeyPath:@"value"];
		_flags.observing = NO;
	}
}

- (BOOL) isItemEditable
{
	return NO;
}

- (unsigned) itemCount
{
	return [(IDLArray*)_type arraySize];
}
- (NSArray*) items
{
	if(!_flags.itemsLoaded){
		NSEnumerator* en = [_value objectEnumerator];
		_items = [[NSMutableArray alloc]init];
		id obj;
		int n = 0;
		while(obj = [en nextObject]){
			IDLItem* item = [IDLItem itemWithIDLType:[(IDLArray*)_type arrayType]];
			[item setItemName:[NSString stringWithFormat:@"%d", n++]];
			[item setItemValue:(obj != [NSNull null] ? obj : nil)];
			[item setEditable:[self isEditable]];
			[_items addObject:item];
		}
		if([_items count] > 0){
			NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
			[_items addObserver:self toObjectsAtIndexes:set forKeyPath:@"itemValue" options:0 context:NULL];
			[_items makeObjectsPerformSelector:@selector(setController:) withObject:[self controller]];
		}
		_flags.itemsLoaded = YES;
	}
	return _items;
}

- itemValue
{
	return [_type name];
}

- (void) setValue:value
{
	if([value isKindOfClass:[NSArray class]])
		value = [IDLArrayObject arrayWithArray:value];
	[super setValue:value];
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)object change:(NSDictionary*)change context:(void*)context
{
	if(object == _value){
		if(_flags.itemsLoaded){
			int changeKind = [[change objectForKey:NSKeyValueChangeKindKey]intValue];
			// No insertion ot removal allowed for array
			if(changeKind == NSKeyValueChangeReplacement){
				NSIndexSet* set = [change objectForKey:NSKeyValueChangeIndexesKey];
				int row = [set firstIndex];
				int i, count = [set count];
				for (i = 0; i < count; i++) {
					id value = [_value objectAtIndex:row];
					[[_items objectAtIndex:row]setValue:value]; // May need to replace item if valuetype ?
				}
				[_controller endEditing];
				[_controller invalidateItem:self];
			}
		}
	}else{
		int n = [_items indexOfObject:object];
		if(n != NSNotFound){
			id itemValue = [object value];
			[_value replaceObjectAtIndex:n withObject:(itemValue? itemValue : [NSNull null])];
		}
	}
}

- (void) _updateItems
{
	NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
	[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
	[self willChangeValueForKey:@"items"];
	[_items release];
	_items = nil;
	_flags.itemsLoaded = NO;
	[self didChangeValueForKey:@"items"];
}

- (void) dealloc
{
	[self stopObservingValueChanges];
	NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
	[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
	[_items release];
	[super dealloc];
}
- (NSString*) valueDescription
{
	return [_value className];
}

@end

@implementation IDLSequenceItem

- (void) startObservingValueChanges
{
	if(!_flags.observing){
		[_value addObserver:self forKeyPath:@"value" options:0 context:NULL];
		_flags.observing = YES;
	}
}
- (void) stopObservingValueChanges
{
	if(_flags.observing){
		[_value removeObserver:self forKeyPath:@"value"];
		_flags.observing = NO;
	}
}

- (BOOL) isItemEditable
{
	return NO;
}

- (unsigned) itemCount
{
	return [_value count];
}
- (unsigned) numberOfItems
{
	return [_items count];
}

- (NSArray*) items
{
	if(!_flags.itemsLoaded){
		NSEnumerator* en = [_value objectEnumerator];
		_items = [[NSMutableArray alloc]init];
		id obj;
		int n = 0;
		while(obj = [en nextObject]){
			IDLItem* item = [IDLItem itemWithIDLType:[(IDLSequence*)_type sequenceType]];
			[item setItemName:[NSString stringWithFormat:@"%d", n++]];
			[item setValue:(obj != [NSNull null] ? obj : nil)];
			[item setEditable:[self isEditable]];
			[_items addObject:item];
		}
		if([_items count] > 0){
			NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
			[_items addObserver:self toObjectsAtIndexes:set forKeyPath:@"itemValue" options:0 context:NULL];
			[_items makeObjectsPerformSelector:@selector(setController:) withObject:[self controller]];
		}
		_flags.itemsLoaded = YES;
	}
	return _items;
}

- itemValue
{
	return [_type name];
}

- (void) setValue:value
{
	if([value isKindOfClass:[NSArray class]])
		value = [IDLArrayObject arrayWithArray:value];
	[super setValue:value];
}

- (void)  updateItemNames
{
	int n = 0;
	NSEnumerator* en = [_items objectEnumerator];
	IDLItem* anItem;
	while(anItem = [en nextObject])
		[anItem setItemName:[NSString stringWithFormat:@"%d", n++]];
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)object change:(NSDictionary*)change context:(void*)context
{
	if(object == _value){
		if(_flags.itemsLoaded){
			int changeKind = [[change objectForKey:NSKeyValueChangeKindKey]intValue];
			switch(changeKind){
				case NSKeyValueChangeInsertion:
					if([_items count] < [_value count]){
						[self willChangeValueForKey:@"numberOfItems"];
						NSIndexSet* set = [change objectForKey:NSKeyValueChangeIndexesKey];
						int row = [set firstIndex];
						int i, count = [set count];
						for (i = 0; i < count; i++) {
							id value = [_value objectAtIndex:row];
							IDLDefinition* type = [(IDLSequence*)[self itemType]sequenceType];
							if([value respondsToSelector:@selector(idlType)])
								type = (IDLDefinition*)[value idlType];
							IDLItem* item = [IDLItem itemWithIDLType:type];
							[item setValue:value];
							[item setEditable:[self isEditable]];
							[item setController:_controller];
							[_items insertObject:item atIndex:row];
							row = [set indexGreaterThanIndex:row];
						}
						[self updateItemNames];
						[_items addObserver:self toObjectsAtIndexes:set forKeyPath:@"itemValue" options:0 context:NULL];
						[self didChangeValueForKey:@"numberOfItems"];
					}
					break;
				case NSKeyValueChangeRemoval:
					if([_items count] > [_value count]){
						[_controller endEditing];
						[self willChangeValueForKey:@"numberOfItems"];
						NSIndexSet* set = [change objectForKey:NSKeyValueChangeIndexesKey];
						[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
						[_items removeObjectsAtIndexes:set];
						[self updateItemNames];
						[self didChangeValueForKey:@"numberOfItems"];
					}
					break;
				case NSKeyValueChangeReplacement: {
					[_controller endEditing];
					NSIndexSet* set = [change objectForKey:NSKeyValueChangeIndexesKey];
					int row = [set firstIndex];
					int i, count = [set count];
					for (i = 0; i < count; i++) {
						id value = [_value objectAtIndex:row];
						[[_items objectAtIndex:row]setValue:value];
					}
					break;
				}
			}
			[_controller invalidateItem:self];
		}
	}else{
		int n = [_items indexOfObject:object];
		if(n != NSNotFound){
			id itemValue = [object value];
			[_value replaceObjectAtIndex:n withObject:(itemValue? itemValue : [NSNull null])];
		}
	}
}

- (void) _updateItems
{
	NSIndexSet* set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
	[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
	[_items release];
	_items = nil;
	_flags.itemsLoaded = NO;
}

- (BOOL) shouldInsertItemWithIDLType:(IDLDefinition*)def
{
	IDLDefinition* seqType = [(IDLSequence*)_type sequenceType];
	return [self isEditable] && [def isDerivedFromIDLType:seqType];
}

- insertNewItemWithIDLType:(IDLDefinition*)def atIndex:(int)index
{
	id itemValue = [def defaultItemValue];
	IDLItem* item = nil;
	if(_flags.itemsLoaded){
		item = [IDLItem itemWithIDLType:def];
		[item setValue:itemValue];
		[item setEditable:[self isEditable]];
		[item setController:[self controller]];
		[_items insertObject:item atIndex:index];
		[item addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
		[self updateItemNames];
	}
	[_value insertObject:itemValue atIndex:index];
	return item;
}

- (BOOL) shouldInsertItem:(IDLItem*)item
{
	IDLDefinition* seqType = [(IDLSequence*)_type sequenceType];
	return [self isEditable] && [[item itemType] isDerivedFromIDLType:seqType];
}

- (void) insertItem:(IDLItem*)item atIndex:(int)index
{
	id itemValue = [item value];
	if(_flags.itemsLoaded){
		[item setEditable:[self isEditable]];
		[item setController:[self controller]];
		[_items insertObject:item atIndex:index];
		[item addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
		[item startObservingValueChanges];
		int n = 0;
		NSEnumerator* en = [_items objectEnumerator];
		while(item = [en nextObject])
			[item setItemName:[NSString stringWithFormat:@"%d", n++]];
	}
	[_value insertObject:itemValue atIndex:index];
}

- (BOOL) shouldInsertNewItem
{
	return [self isEditable];
}
- (BOOL) shouldRemoveItem
{
	return [self isEditable];
}

- insertNewItemAtIndex:(int)index
{
	IDLDefinition* seqType = [(IDLSequence*)_type sequenceType];
	return [self insertNewItemWithIDLType:seqType atIndex:index];
}

- (void) removeItemAtIndex:(int)index
{
	if(index >= 0 && index < [_items count]){
		NSIndexSet* set = [NSIndexSet indexSetWithIndex:index];
		[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
		[_items removeObjectAtIndex:index];
		[_value removeObjectAtIndex:index];
		[self updateItemNames];
	}
}

- (NSString *) valueDescription
{
	return [_value className];
}

- (void) dealloc
{
	[self stopObservingValueChanges];
	NSIndexSet *set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
	[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
	[_items release];
	[super dealloc];
}
@end

@implementation IDLBooleanItem
- (NSCell *) dataCell
{
	NSButtonCell *cell = [[NSButtonCell alloc]initTextCell:@""];
	[cell setButtonType:NSSwitchButton];
	[cell setControlSize:NSSmallControlSize];
	[cell setEnabled:[self isEditable]];
	return [cell autorelease];
}

- (BOOL) isItemEditable
{
	return NO;
}
@end

@implementation IDLOctetItem


@end

@implementation IDLNumberItem
- (void)resetCell:(NSTextFieldCell *)cell
{
	NSNumberFormatter *formatter = [[NSNumberFormatter alloc]init];
	[cell setFormatter:nil];
	switch([_type kind]){
		case tk_ulonglong:
		case tk_ulong:
		case tk_ushort:
			[formatter setMinimum:[NSNumber numberWithLong:0]];
		case tk_longlong:
		case tk_long:
		case tk_short:
			[formatter setFormat:@"0"];
			[formatter setAllowsFloats:NO];
			break;
		case tk_float:
		case tk_double:
			[formatter setAllowsFloats:YES];
			[formatter setFormat:@"0.###################"];
			break;
	}
	[cell setTextColor:[NSColor blackColor]];
	[cell setPlaceholderString:[_type name]];
	[cell setFormatter:formatter];
	[formatter release];
}

@end

@implementation IDLCharItem
- (void)resetCell:(NSTextFieldCell *)cell
{
	IDLCharFormatter *formatter = [[IDLCharFormatter alloc]init];
	[cell setFormatter:nil];
	[cell setTextColor:[NSColor blackColor]];
	[cell setPlaceholderString:[_type name]];
	[cell setFormatter:formatter];
	[formatter release];
}

@end

@implementation IDLStringItem

- (void)resetCell:(NSTextFieldCell *)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[NSColor blackColor]];
	[cell setPlaceholderString:[_type name]];
}

@end

@implementation IDLFunctionItem

- (BOOL) isItemEditable
{
	return NO;
}

- (unsigned) itemCount
{
	switch([[(IDLFunction *)_type returnType]kind]){
		default:
			return [[(IDLFunction *)_type arguments]count] + 1;
		case tk_void:
			return [[(IDLFunction *)_type arguments]count];
	}
}
- (NSArray *) items
{
	if(!_flags.itemsLoaded){
		_items = [[(IDLFunction *)_type itemArrayForItem:self]retain];
		if([_items count] > 0){
			NSIndexSet *set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
			[_items addObserver:self toObjectsAtIndexes:set forKeyPath:@"itemValue" 
				options:NSKeyValueObservingOptionNew|NSKeyValueObservingOptionOld context:NULL];
			[_items makeObjectsPerformSelector:@selector(setController:) withObject:[self controller]];
		}
		_flags.itemsLoaded = YES;
	}
	return _items;
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
	[_value setValue:[object value] forKey:[object itemName]];
}

- itemValue
{
	return [_type name];
}

- (void) _updateItems
{
	NSEnumerator* en = [_items objectEnumerator];
	IDLItem *item;
	while(item = [en nextObject])
		[item setValue:[_value objectForKey:[item itemName]]];
	[_controller invalidateItem:self];
}

- (void) dealloc
{
	NSIndexSet *set = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_items count])];
	[_items removeObserver:self fromObjectsAtIndexes:set forKeyPath:@"itemValue"];
	[_items release];
	[super dealloc];
}

@end


@implementation IDLEnumItem

- (BOOL) isItemEditable
{
	return NO;
}

@end


@implementation IDLIORItem

- (NSCell *) dataCell
{
	NSCell *cell = [[IORCell alloc]init];
	[cell setEditable:[self isEditable]];
	return [cell autorelease];
}

- (BOOL) isItemEditable
{
	return NO;
}

- (BOOL) shouldSetIDLType:(IDLDefinition *)type
{
	return [self isEditable] && [type isDerivedFromIDLType:[self itemType]];
}

- (void) setIDLType:(IDLDefinition *)def
{
	// This method is called from within a user action
	// Needs to notify observers
	[self setItemValue:nil];
}

- (void) takeValueFromItem:(IDLItem *)item
{
	id value = [item value];
	if(value == nil || [value isKindOfClass:[CORBAObject class]])
		// This method is called from within a user action
		// Needs to notify observers
		[self setItemValue:[item value]];
}

- (NSString *) valueDescription
{
	if(_value)
		return [(CORBAObject *)_value toString];
	else
		return @"<nil>";
}

extern NSString *CORBAIORPboardType;

- (BOOL) writeToPasteboard:(NSPasteboard *)pboard
{
	NSArray *types = _value
		?[NSArray arrayWithObjects:IDLItemPboardType, CORBAIORPboardType, NSStringPboardType, nil]
		:[NSArray arrayWithObject:IDLItemPboardType];
	NSData *data = [NSKeyedArchiver archivedDataWithRootObject:self];
	[pboard declareTypes:types owner:nil];
	[pboard setData:data forType:IDLItemPboardType];
	if(_value){
		NSString *ior = [_value toString];
		[pboard setString:ior forType:CORBAIORPboardType];
		[pboard setString:ior forType:NSStringPboardType];
	}
	return YES;
}

static BOOL takeValueFromPasteboard(id self, NSPasteboard *pboard, NSString *type)
{
    if([[self controller]respondsToSelector:@selector(setScope)])
        [[self controller]setScope];
    if([type isEqualToString:CORBAIORPboardType]){
        NSString *ior = [pboard stringForType:CORBAIORPboardType];
        id obj = nil;
        @try{
			obj = [CORBAObject objectWithIOR:ior];
			if(![obj interface])
				[obj narrowInterface];
			if([[obj interface]adoptsInterface:(IDLInterface *)[self itemType]])
				[self setItemValue:obj];
			else
				obj = nil;
        }@catch(NSException *localException){
			NSBeginAlertSheet(@"Exception", NULL,NULL,NULL,
				[[self controller]window], nil, NULL, NULL, NULL,
				@"Invalid CORBA URL - %@: %@", [localException name], [localException reason]);
			obj = nil;
        }
		return obj != nil;
    }else
    if([type isEqualToString:NSStringPboardType]){
        NSString *ior = [pboard stringForType:NSStringPboardType];
        id obj = nil;
        @try{
			obj = [CORBAObject objectWithURL:ior];
			if([[obj interface]adoptsInterface:(IDLInterface *)[self itemType]])
				[self setItemValue:obj];
			else
				obj = nil;
        }@catch(NSException *localException){
			NSBeginAlertSheet(@"Exception", NULL,NULL,NULL,
				[[self controller]window], nil, NULL, NULL, NULL,
				@"Invalid CORBA URL - %@: %@", [localException name], [localException reason]);
			obj = nil;
        }
		return obj != nil;
    }
	return NO;
}

- (BOOL) takeValueFromPasteboard:(NSPasteboard *)pboard forType:(NSString *)type
{
	return takeValueFromPasteboard(self, pboard, type);
}

@end

@implementation IDLValueItem

- initWithIDLType:(IDLDefinition *)def
{
	self = [super initWithIDLType:def];
	_value = nil;
	_valueType = [def retain];
	_valueItem = [(IDLValueType *)_valueType createValueItemWithName:[self itemName]];
	[_valueItem setEditable:[self isEditable]];
	[_valueItem setIsInTree:NO];
	[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
	[_valueItem addObserver:self forKeyPath:@"numberOfItems" options:0 context:NULL];
	return self;
}

- (void) setEditable:(BOOL)flag
{
	[super setEditable:flag];
	[_valueItem setEditable:[self isEditable]];
}
	
- (void) dealloc
{
	[self stopObservingValueChanges];
	[self setItemType:nil];
	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem removeObserver:self forKeyPath:@"numberOfItems"];
	[_valueItem release];
	[_valueType release];
	[super dealloc];
}

- (NSCell *) dataCell
{
	if(_value)
		return [_valueItem dataCell];
	else
		return nil;
}
- (void)resetCell:(NSTextFieldCell *)cell
{
	if(_valueItem)
		[_valueItem resetCell:cell];
	else
		[_valueType resetItemCell:cell];
	[cell setPlaceholderString:@"<nil>"];
}

- (BOOL) isLeaf
{
	if(_value)
		return [_valueItem isLeaf];
	else
		return YES;
}

- itemValue
{
	if(_value)
		return [_valueItem itemValue];
	else
		return nil;
}

- (IDLDefinition*)itemType
{
	return _valueType; // TEST
}

- (void) setItemType:(IDLDefinition*)type
{
	[self willChangeValueForKey:@"itemType"];
	[_valueType release];
	_valueType = [type retain];
	[self didChangeValueForKey:@"itemType"];
}
	
// value's idl type may be either valuetype or interface
- (void) _updateItems
{
	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem removeObserver:self forKeyPath:@"numberOfItems"];
	[_valueItem release];
	_valueItem = [(IDLValueType*)_valueType createValueItemWithName:@"value"];
	[_valueItem setEditable:[self isEditable]];
	[_valueItem setItemValue:_value];
	[_valueItem setController:_controller];
	[_valueItem setIsInTree:NO];
	[_valueItem setEditable:[self isEditable]];
	[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
	[_valueItem addObserver:self forKeyPath:@"numberOfItems" options:0 context:NULL];
	[_controller invalidateItem:self];
}

- (void) setValue:value
{
	BOOL needsReload = NO;
	if([value respondsToSelector:@selector(idlType)]){
		IDLDefinition* type = (IDLDefinition*)[value idlType];
		// In the case _type is a valuebox the 'type' may be IDLStruct or IDLAny
		if([type isDerivedFromIDLType:_type] && _valueType != type){
			[self setItemType:type];
			needsReload = YES;
		}
	}
	if([value respondsToSelector:@selector(wrappedValue)])
		value = [value wrappedValue];
	if(value != _value){
		[_value release];
		_value = [value retain];
		// _value may have a valuebox's value type ?
		// make sure it has proper idl type, so it can be marshaled
		if([_value respondsToSelector:@selector(setIDLType:)])
			[_value setIDLType:_valueType];
		if(needsReload)
			[self _updateItems];
		else{
			[_valueItem setValue:_value];
			[_controller invalidateItem:self];
		}
	}else if(needsReload)
		[self _updateItems];
}

// Returns _value if it can be marshalled or a wrapper otherwise
- value
{
	if(_value){
		if([_value respondsToSelector:@selector(valuetype)]) // idlType ?
			return _value;
		else
			return [_type wrapperValueForValue:_value withIDLType:_valueType];
	}else
		return nil;
}
			
- (BOOL) isItemEditable
{
	return [_valueItem isItemEditable];
}

- (void) setController:ctrl
{
	_controller = ctrl;
	[_valueItem setController:ctrl];
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)item change:(NSDictionary*)change context:(void*)context
{
	if(item == _valueItem){
		if([keyPath isEqualToString:@"itemValue"])
			[self setItemValue:[item value]];
		else if([keyPath isEqualToString:@"numberOfItems"])
			[_controller invalidateItem:self];
	}
}

- (BOOL) shouldSetIDLType:(IDLDefinition*)type
{
	return [type isDerivedFromIDLType:_type];
}

// Sets the value idl type to def
// Sets the value to the default value for def
// Expands self
// This method is called when a definition is dragged in
- (void) setIDLType:(IDLDefinition*)def
{
	if([def isDerivedFromIDLType:_type]){
		[self setItemType:def];
		[self setItemValue:[def defaultValueItemValue]];
		[self _updateItems];
		[_controller invalidateItem:self];
		[_controller expandItem:self];
	}
}

// Return YES if either item is suitable for self or for _valueItem.
// The latter case can take place for a valuebox
- (BOOL) shouldTakeValueFromItem:(IDLItem*)item
{
	return [self isEditable] && 
		([[item itemType]isDerivedFromIDLType:_type] || 
		 [[item itemType]isDerivedFromIDLType:[_valueItem itemType]]);
}

// Overrides superclass's method
// For valuebox, items of _valueItem's type can be dragged in or pasted
- (void) takeValueFromItem:(IDLItem*)item
{
	if([[item itemType]isDerivedFromIDLType:_type]){
		BOOL changeType = (_valueType != [item itemType]);
		if(changeType){
			[self setItemType:[item itemType]];
		}
		[self setItemValue:[item value]];
		// reload self if type has changed
		if(changeType){
			[self _updateItems];
			[_controller invalidateItem:self];
		}
	}else if([[item itemType]isDerivedFromIDLType:[_valueItem itemType]])
		// for valueboxes
		[_valueItem setValue:[item value] forKey:@"itemValue"];
}

// Methods for inserting/removing items for the case of valuebox
// when valuebox's type is sequence
- (BOOL) shouldInsertItem:(IDLItem*)item
{
	return [_valueItem shouldInsertItem:item];
}
- (BOOL) shouldInsertItemWithIDLType:(IDLDefinition*)def
{
	return [_valueItem shouldInsertItemWithIDLType:def];
}
- insertNewItemWithIDLType:(IDLDefinition*)def atIndex:(int)index
{
	return [(IDLSequenceItem*)_valueItem insertNewItemWithIDLType:def atIndex:index];
}
- (void) insertItem:(IDLItem*)item atIndex:(int)index
{
	[_valueItem insertItem:item atIndex:index];
}
- (BOOL) shouldInsertNewItem
{
	return _value && [_valueItem respondsToSelector:_cmd] && [(IDLSequenceItem*)_valueItem shouldInsertNewItem];
}

- (BOOL) shouldRemoveItem
{
	return _value && [_valueItem respondsToSelector:_cmd] && [(IDLSequenceItem*)_valueItem shouldRemoveItem];
}

- insertNewItemAtIndex:(int)index
{
	return [(IDLSequenceItem*)_valueItem insertNewItemAtIndex:index];
}
- (void) removeItemAtIndex:(int)index
{
	[(IDLSequenceItem*)_valueItem removeItemAtIndex:index];
}

// valuetype shares item (cell) with _support, is it value class or any other type
- (unsigned) itemCount
{
	return [_valueItem itemCount];
}
- (NSArray*) items
{
	return [_valueItem items];
}

// Resets value idl type to self's idl type
// and set the value to the default for that idl type
// Expands item
- (void) resetItemValue
{
	if([self isEditable])
		[self setIDLType:_type];
}

// Same as previous, but does it only if value's value is nil
// Called when item is doble-clicked
- (void) setItemValue
{
	if([self isEditable] && _value == nil)
		[self setIDLType:_type];
}

// Resets value's idl type and sets value's value to nil
- (void) clearItemValue
{
	if([self isEditable]){
		[self setItemType:_type];
		[self setItemValue:nil];
		[self _updateItems];
		[_controller invalidateItem:self];
	}
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	[super encodeWithCoder:coder];
	[coder encodeObject:_valueItem forKey:@"valueItem"];
	[coder encodeObject:_valueType forKey:@"valueType"];
}

- initWithCoder:(NSCoder*)coder
{
	self = [super initWithCoder:coder];
	_valueItem = [[coder decodeObjectForKey:@"valueItem"]retain];
	_valueType = [[coder decodeObjectForKey:@"valueType"]retain];
	[_valueItem setIsInTree:NO];
	[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
	[_valueItem addObserver:self forKeyPath:@"numberOfItems" options:0 context:NULL];
	return self;
}

@end

@implementation IDLAbstractItem

- (BOOL) takeValueFromPasteboard:(NSPasteboard*)pboard forType:(NSString*)type
{
	return takeValueFromPasteboard(self, pboard, type);
}

- value
{
	if([_value isKindOfClass:[CORBAObject class]])
		return _value;
	else
		return [super value];
}

- (void) setValue:value
{
	[self stopObservingValueChanges];
	if([value isKindOfClass:[CORBAObject class]]){
		if(value != _value){
			if([value interface] == nil)
				[value narrowInterface];
			[self setItemType:[value interface]];
			[_value release];
			_value = [value retain];
			[_valueItem removeObserver:self forKeyPath:@"itemValue"];
			[_valueItem removeObserver:self forKeyPath:@"numberOfItems"];
			[_valueItem release];
			_valueItem = [_valueType createIDLItemWithName:@"value"];
			[_valueItem setItemValue:_value];
			[_valueItem setEditable:[self isEditable]];
			[_valueItem setController:_controller];
			[_valueItem setIsInTree:NO];
			[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
			[_valueItem addObserver:self forKeyPath:@"numberOfItems" options:0 context:NULL];
			[_controller invalidateItem:self];
		}
	}else
		[super setValue:value];
	[self startObservingValueChanges];
}

@end

@implementation IDLAnyItem
- (BOOL) isItemEditable
{
	return NO;
}

- (unsigned) itemCount
{
	return [_value idlType]? 1 : 0;
}

- itemValue
{
	NSString* value= [[_value idlType]name];
	return value? value : @"any";
}

- (NSArray*) items
{
	if(!_flags.itemsLoaded){
		IDLDefinition* valueType = (IDLDefinition*)[_value idlType];
		if(valueType){
			_valueItem = [valueType createIDLItemWithName:@"any"];
			[_valueItem setValue:[_value valueForKey:@"value"]];
			[_valueItem setEditable:[self isEditable]];
			[_valueItem setController:[self controller]];
			[_valueItem addObserver:self forKeyPath:@"itemValue" 
				options:NSKeyValueObservingOptionNew|NSKeyValueObservingOptionOld context:NULL];
		}
		_flags.itemsLoaded = YES;
	}
	if(_valueItem)
		return [NSArray arrayWithObject:_valueItem];
	else
		return nil;
}

- (void) startObservingValueChanges
{
	if(!_flags.observing){
		[_value addObserver:self forKeyPath:@"value" options:0 context:NULL];
		[_value addObserver:self forKeyPath:@"idlType" options:0 context:NULL];
		_flags.observing = YES;
	}
}

- (void) stopObservingValueChanges
{
	if(_flags.observing){
		[_value removeObserver:self forKeyPath:@"value"];
		[_value removeObserver:self forKeyPath:@"idlType"];
		_flags.observing = NO;
	}
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)object change:(NSDictionary*)change context:(void*)context
{
	if(object == _valueItem){
		[_value setValue:[object value] forKey:@"value"];
		//[_controller invalidateItem:self];
	}else if(object == _value){
		if([keyPath isEqualToString:@"value"]){
			[_valueItem setValue:[_value value]];
			[_controller endEditing];
			[_controller invalidateItem:self];
		}else if([keyPath isEqualToString:@"idlType"])
			[self _updateItems];
	}
}

- (BOOL) shouldSetIDLType:(IDLDefinition*)type
{
	return [self isEditable];
}

- (void) _updateItems
{
	_flags.itemsLoaded = NO;
	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem release];
	_valueItem = nil;
}

- (void) setIDLType:(IDLDefinition*)def
{
	[_value setIDLType:def];
	// [self _updateItems]; // this will be done in observing method
	[_value setValue:[def defaultItemValue]];
	[_controller invalidateItem:self];
}

- (BOOL) shouldTakeValueFromItem:(IDLItem*)item
{
	return [self isEditable];
}

- (void) takeValueFromItem:(IDLItem*)item
{
	[self _updateItems];
	if([[item itemType]kind] == tk_any){
		[_value setIDLType:[[item value]idlType]];
		[_value setValue:[[item value]value]];
	}else{
		[_value setIDLType:[item itemType]];
		[_value setValue:[item value]];
	}
}

- (NSString*) valueDescription
{
	if(_value)
		return [_value description];
	else
		return @"<null>";
}

- (void) dealloc
{
	[self stopObservingValueChanges];
	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem release];
	[super dealloc];
}
@end


@implementation IDLUnionItem

- (BOOL) isItemEditable
{
	return NO;
}

- (unsigned) itemCount
{
	if(_flags.itemsLoaded)
		return _valueItem? 2 : 1;
	else
		return [[self items]count];
}
- (unsigned) numberOfItems
{
	return _valueItem? 2 : 1;
}

- (NSArray*) items
{
	if(!_flags.itemsLoaded){
		IDLDefinition* switchType = [(IDLUnion*)_type switchType];
		_descriptorItem = [switchType createIDLItemWithName:@"descriptor"];
		[_descriptorItem setValue:[_value valueForKey:@"descriptor"]];
		[_descriptorItem setEditable:[self isEditable]];
		[_descriptorItem setController:[self controller]];
		IDLSwitchCase* caseObject = [(IDLUnion*)_type caseForValue:[_descriptorItem value]];
		_valueItem = [[caseObject caseType]createIDLItemWithName:[caseObject name]];
		[_valueItem setValue:[_value valueForKey:@"value"]];
		[_valueItem setEditable:[self isEditable]];
		[_valueItem setController:[self controller]];
		[_descriptorItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
		[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
		_flags.itemsLoaded = YES;
	}
	return [NSArray arrayWithObjects:_descriptorItem, _valueItem, nil];
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)object change:(NSDictionary*)change context:(void*)context
{
	if(object == _descriptorItem){
		IDLSwitchCase* caseObject = [(IDLUnion*)_type caseForValue:[object value]];
		id caseValue = [[caseObject caseType] defaultItemValue];
		[self willChangeValueForKey:@"numberOfItems"]; // if in valuetype
		[_valueItem removeObserver:self forKeyPath:@"itemValue"];
		[_valueItem release];
		_valueItem = [[caseObject caseType]createIDLItemWithName:[caseObject name]];
		[_valueItem setValue:caseValue];
		[_valueItem setEditable:[self isEditable]];
		[_value setValue:[object value] forKey:@"descriptor"];
		[_value setValue:caseValue forKey:@"value"];
		[_valueItem setController:_controller];
		[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
		[self didChangeValueForKey:@"numberOfItems"];
		[_controller invalidateItem:self];
	}else if(object == _valueItem)
		[_value setValue:[object value] forKey:@"value"];
	else if(object == _value){
		if(_flags.itemsLoaded){
			id descValue = [_value valueForKey:@"descriptor"];
			if([_descriptorItem value] != descValue){
				[_descriptorItem setValue:descValue];
				[self willChangeValueForKey:@"numberOfItems"]; // if in valuetype
				IDLSwitchCase* caseObject = [(IDLUnion*)_type caseForValue:descValue];
				[_valueItem removeObserver:self forKeyPath:@"itemValue"];
				[_valueItem release];
				_valueItem = [[caseObject caseType]createIDLItemWithName:[caseObject name]];
				[_valueItem setValue:[_value valueForKey:@"value"]];
				[_valueItem setEditable:[self isEditable]];
				[_valueItem setController:_controller];
				[_valueItem addObserver:self forKeyPath:@"itemValue" options:0 context:NULL];
				[self didChangeValueForKey:@"numberOfItems"];
				[_controller endEditing];
				[_controller invalidateItem:self];
			}else{
				[_valueItem setValue:[_value valueForKey:@"value"]];
				[_controller endEditing];
				[_controller invalidateItem:_valueItem];
			}
		}
	}
}

- (void) startObservingValueChanges
{
	if(!_flags.observing){
		//[_value addObserver:self forKeyPath:@"descriptor" options:0 context:NULL];
		[_value addObserver:self forKeyPath:@"value" options:0 context:NULL];
		_flags.observing = YES;
	}
}

- (void) stopObservingValueChanges
{
	if(_flags.observing){
		//[_value removeObserver:self forKeyPath:@"descriptor"];
		[_value removeObserver:self forKeyPath:@"value"];
		_flags.observing = NO;
	}
}

- itemValue
{
	return [_type name];
}

- (void) _updateItems
{
	[_descriptorItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_descriptorItem release];
	[_valueItem release];
	_descriptorItem = nil;
	_valueItem = nil;
	_flags.itemsLoaded = NO;
}

- (NSString*) valueDescription
{
	return [_value description];
}

- (void) dealloc
{
	[self stopObservingValueChanges];
	[_descriptorItem removeObserver:self forKeyPath:@"itemValue"];
	[_valueItem removeObserver:self forKeyPath:@"itemValue"];
	[_descriptorItem release];
	[_valueItem release];
	[super dealloc];
}

@end


@implementation IDLEnumCell

- init
{
	self = [super initTextCell:@"" pullsDown:NO];
	[self removeAllItems];
	[self addItemWithTitle:@""];
	[[self itemAtIndex:0]setRepresentedObject:nil];
	[[self itemAtIndex:0]setTag:-1];
	[self setBezeled:NO];
	[self setBordered:NO];
	[self setControlSize:NSSmallControlSize];
	[self setFont:[NSFont systemFontOfSize:[NSFont smallSystemFontSize]]];
	[self setAutoenablesItems:YES];
	return self;
}

- objectValue
{
	return [[self selectedItem]representedObject];
}

- (void) setObjectValue:value
{
	if([value isKindOfClass:[NSNumber class]]){
		[self selectItemWithTag:[value intValue]];
	}else
	if(value == nil)
		[self selectItemAtIndex:0];
}

- copyWithZone:(NSZone*)zone
{
	return [self retain];
}

- (BOOL)validateMenuItem:(NSMenuItem*)menuItem
{
	return [self isEditable];
}

@end

@implementation IORCell
- init
{
	self = [super initTextCell:@""];
	[self setBezeled:NO];
	[self setBordered:NO];
	[self setControlSize:NSSmallControlSize];
	[self setFont:[NSFont systemFontOfSize:11.0]];
	[self setPlaceholderString:@"IOR"];
	[self setFormatter:[[[IORFormatter alloc]init]autorelease]];
	return self;
}

@end

@implementation CORBAObject(IDLItem)
- (void) encodeWithCoder:(NSCoder*)coder
{
	NSString* ior = [self toString];
	[coder encodeObject:ior forKey:@"ior"];
}

- initWithCoder:(NSCoder*)coder
{
	self = [super init];
	NSString* ior = [coder decodeObjectForKey:@"ior"];
	[self release];
	return [[CORBAObject objectWithIOR:ior]retain];
}

- copyWithZone:(NSZone*)zone
{
	return [self retain];
}

@end

@implementation IORFormatter
- (NSAttributedString*)attributedStringForObjectValue:(id)object withDefaultAttributes:(NSDictionary*)attributes
{
	if([object isKindOfClass:[CORBAObject class]]){
		NSString* s = @"IOR";
		switch(_mode){
			case 0:
				s = [NSString stringWithFormat:@"IOR:%@:%u/%@", 
					[object host], [[(CORBAObject*)object port]unsignedShortValue], [object objectID]];
				break;
			case 1:
			default:
				s = [NSString stringWithFormat:@"IOR:%@:%u", 
					[object host], [[(CORBAObject*)object port]unsignedShortValue]];
				break;
		}
		NSMutableAttributedString* string = [[NSMutableAttributedString alloc]initWithString:s attributes:attributes];
		NSDictionary* at = [NSDictionary dictionaryWithObject:[NSFont boldSystemFontOfSize:11.0] forKey:NSFontAttributeName];
		[string setAttributes:at range:NSMakeRange(0,4)];
		return [string autorelease];
	}else
		return nil;
}
- (NSString *)stringForObjectValue:(id)object
{
	if([object isKindOfClass:[CORBAObject class]])
		return [NSString stringWithFormat:@"IOR:%@:%u/%@", 
			[object host], [[(CORBAObject*)object port]unsignedShortValue], [object objectID]];
	else
		return nil;
}
- (BOOL)getObjectValue:(id*)obj forString:(NSString*)string errorDescription:(NSString**)error
{
	BOOL ok = YES;
	if([string length] > 0){
		@try{
			*obj = [CORBAObject objectWithURL:string];
		}@catch(NSException *localException){
			if(error)
				*error = [NSString stringWithFormat:@"Invalid CORBA URL - %@: %@", 
					[localException name], [localException reason]];
			ok = NO;
		}
	}else
		*obj = nil;
	return ok;
}

- (int) formatMode
{
	return _mode;
}

- (void) setFormatMode:(int)mode
{
	_mode = mode;
}

@end

@implementation IDLCharFormatter

- (NSString *)stringForObjectValue:(id)object
{
	if([object isKindOfClass:[NSNumber class]]){
		unichar c = [object unsignedCharValue];
		if(c > ' ' && c < 128)
			return [NSString stringWithCharacters:&c length:1];
		else
		switch(c){
			case '\t': return @"\\t";
			case '\n': return @"\\n";
			case '\r': return @"\\r";
			default: {
				unichar s[4] = {'\\', 0, 0, 0};
				int n = 1;
				if(s[n] = c>>6)
					n += 1;
				s[n] = (c>>3) & 0x07;
				if(s[n] != 0 || n > 1)
					n += 1;
				s[n++] = c & 0x07;
				s[1] += '0'; s[2] += '0'; s[3] += '0';
				return [NSString stringWithCharacters:s length:n];
			}
		}
	}else
		return nil;
}

- (BOOL)getObjectValue:(id*)obj forString:(NSString*)string errorDescription:(NSString**)error
{
	if([string length] > 0 && [string characterAtIndex:0] == '\\'){
		NSScanner* scanner = [NSScanner scannerWithString:[string substringFromIndex:1]];
		[scanner setCharactersToBeSkipped:[NSCharacterSet characterSetWithCharactersInString:@""]];
		NSString* result;
		if([scanner scanCharactersFromSet:[NSCharacterSet characterSetWithCharactersInString:@"01234567"] intoString:&result]){
			if([scanner isAtEnd] && [result length] < 4){
				int n;
				unsigned s = 0;
				for(n = 0; n<[result length]; n++)
					s = s*8 + [result characterAtIndex:n] - '0';
				if(s < 256){
					*obj = [NSNumber numberWithUnsignedChar:s];
					return YES;
				}
			}
		}else
		if([scanner scanCharactersFromSet:[NSCharacterSet characterSetWithCharactersInString:@"tnr"] intoString:&result]){
			if([scanner isAtEnd] && [result length] == 1){
				switch([result characterAtIndex:0]){
					case 't':
						*obj = [NSNumber numberWithUnsignedChar:'\t'];
						return YES;
					case 'n':
						*obj = [NSNumber numberWithUnsignedChar:'\n'];
						return YES;
					case 'r':
						*obj = [NSNumber numberWithUnsignedChar:'\r'];
						return YES;
				}
			}
		}
	}else if([string length] == 1){
		*obj = [NSNumber numberWithUnsignedChar:[string characterAtIndex:0]];
		return YES;
	}
	if(error)
		*error = @"Not a character";
	return NO;
}

@end
