/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of IDLBrowser.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#import "IDLItem.h"
#import "IDLFunction.h"
#import "IDLInterface_Private.h"
#import "IDLPrimitive.h"
#import "IDLStruct.h"
#import "IDLStructMember.h"
#import "IDLEnum.h"
#import "IDLOctetArray.h"
#import "IDLOctetSequence.h"
#import "IDLArgument.h"
#import "IDLUnion.h"
#import "IDLSwitchCase.h"
#import "IDLAny.h"
#import "IDLValueType_Private.h"
#import "IDLValueClass.h"
#import "IDLValueMember.h"
#import "IDLTypedef.h"
#import "IDLAbstractBase.h"
#import "IDLTCKind.h"
#import "IDLObject.h"
#import "IDLValueObject.h"
#import "OctetArrayFormatter.h"


@implementation IDLDefinition(IDLItem)
- (Class) itemClass
{
	return [IDLItem class];
}

- (BOOL) isLeaf
{
	return YES;
}

- (BOOL) isItemEditable
{
	return NO;
}

- (IDLItem*) createIDLItemWithName:(NSString*)name
{
	IDLItem* item = [[[self itemClass]alloc]initWithIDLType:self];
	[item setItemName:name];
	return item;
}
- (IDLItem*) createIDLItemWithName:(NSString*)name forIDLType:(IDLDefinition*)type
{
	return [self createIDLItemWithName:name];
}

- defaultItemValue
{
	return nil;
}
- defaultValueItemValue
{
	[self doesNotRecognizeSelector:_cmd];
	return nil;
}
- defaultItemValueForIDLType:(IDLDefinition*)type
{
	return [self defaultItemValue];
}

// If a value cannot be marshaled it needs a wrapper value
// This methos is overriden by IDLValueType and by IDLInterface
- wrapperValueForValue:value withIDLType:(IDLDefinition*)type
{
	return value;
}

- (NSCell*) itemDataCell
{
	return nil;
}
- (void)resetItemCell:(NSTextFieldCell*)cell
{
	[cell setPlaceholderString:[self name]];
}

@end

@implementation IDLAttribute(IDLItem)
- (Class) itemClass
{
	return [IDLAttributeItem class];
}

- (BOOL) isLeaf
{
	return [_type isLeaf];
}

- (IDLItem*) createValueItemWithName:(NSString*)name
{
	return [_type createIDLItemWithName:name];
}

- defaultItemValue
{
	return [_type defaultItemValue];
}

- (NSCell*) itemDataCell
{
	return [_type itemDataCell];
}
- (void)resetItemCell:(NSTextFieldCell*)cell
{
	[_type resetItemCell:cell];
}

- (NSMethodSignature*) methodSignatureForSelector:(SEL)sel
{
	if([IDLAttribute instancesRespondToSelector:sel])
		return [super methodSignatureForSelector:sel];
	else
		return [_type methodSignatureForSelector:sel];
}
- (void) forwardInvocation:(NSInvocation*)inv
{
	[inv invokeWithTarget:_type];
}

- (NSString*) commandName
{
	return @"Get";
}

@end

@implementation IDLArray(IDLItem)
- (Class) itemClass
{
	return [IDLArrayItem class];
}
- (BOOL) isLeaf
{
	return NO;
}

- defaultItemValue
{
	IDLArrayObject* array = [IDLArrayObject array];
	int count = [self arraySize];
	while(count-- > 0){
		id obj = [_type defaultItemValue];
		[array addObject:(obj? obj : [NSNull null])];
	}
	return array;
}
- (void) resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}

@end

@implementation IDLSequence(IDLItem)
- (Class) itemClass
{
	return [IDLSequenceItem class];
}
- (BOOL) isLeaf
{
	return NO;
}
- defaultItemValue
{
	IDLArrayObject* array = [IDLArrayObject array];
	int count = _frameCount;
	while(count-- > 0){
		id obj = [type defaultItemValue];
		[array addObject:(obj? obj : [NSNull null])];
	}
	return array;
}
- (void) resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}
@end

@implementation IDLStruct(IDLItem)
- (Class) itemClass
{
	return [IDLStructItem class];
}
- (BOOL) isLeaf
{
	return NO;
}

- defaultItemValue
{
	IDLObject* obj = [IDLObject objectWithIDLType:self];
	NSEnumerator* en = [_members objectEnumerator];
	IDLStructMember* member;
	while(member = [en nextObject]){
		IDLDefinition* type = [member type];
		NSString* name = [member name];
		[obj setValue:[type defaultItemValue] forKey:name];
	}
	return obj;
}
- (NSMutableArray*) itemArrayForItem:(IDLItem*)structItem
{
	NSEnumerator* en = [_members objectEnumerator];
	IDLStructMember* member;
	NSMutableArray* array = [NSMutableArray array];
	id value = [structItem value];
	while(member = [en nextObject]){
		IDLDefinition* type = [member type];
		NSString* name = [member name];
		IDLItem* item = [IDLItem itemWithIDLType:type];
		[item setItemName:name];
		[item setValue:[value valueForKey:name]];
		[item setEditable:[structItem isEditable]];
		[item setController:[structItem controller]];
		[array addObject:item];
	}
	return array;
}

- (void) removeValueObserver:(IDLItem*)structItem forValue:value
{
	NSEnumerator* en = [_members objectEnumerator];
	IDLStructMember* member;
	while(member = [en nextObject]){
		NSString* name = [member name];
		[value removeObserver:structItem forKeyPath:name];
	}
}

- (void) addValueObserver:(IDLItem*)structItem forValue:value
{
	NSEnumerator* en = [_members objectEnumerator];
	IDLStructMember* member;
	while(member = [en nextObject]){
		NSString* name = [member name];
		[value addObserver:structItem forKeyPath:name options:0 context:NULL];
	}
}

- (void) resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}

@end

@implementation IDLFunction(IDLItem)
- (Class) itemClass
{
	return [IDLFunctionItem class];
}
- (BOOL) isLeaf
{
	return NO;
}

- defaultItemValue
{
	NSMutableDictionary* dict = [NSMutableDictionary dictionary];
	NSEnumerator* en = [_args objectEnumerator];
	IDLArgument* arg;
	while(arg = [en nextObject]){
		IDLDefinition* type = [arg definition];
		NSString* name = [arg name];
		[dict setValue:[type defaultItemValue] forKey:name];
	}
	return dict;
}
- (NSMutableArray*) itemArrayForItem:(IDLItem*)funcItem
{
	NSEnumerator* en = [_args objectEnumerator];
	IDLArgument* arg;
	NSMutableArray* array = [NSMutableArray array];
	id value = [funcItem value];
	while(arg = [en nextObject]){
		IDLDefinition* type = [arg definition];
		NSString* name = [arg name];
		IDLItem* item = [IDLItem itemWithIDLType:type];
		[item setItemName:name];
		[item setItemValue:[value objectForKey:name]];
		[item setEditable:[arg isInMode]];
		[array addObject:item];
	}
	if([_returnType kind] != tk_void){
		IDLItem* item = [IDLItem itemWithIDLType:_returnType];
		[item setItemName:@"$return"];
		[item setItemValue:[value objectForKey:@"$return"]];
		[item setEditable:NO];
		[array addObject:item];
	}
	return array;
}
- (void) resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}

- (NSString*) commandName
{
	return @"Send";
}

@end

@implementation IDLPrimitive(IDLItem)
- (Class) itemClass
{
	switch([self kind]){
		default:
			return [IDLNumberItem class];
		case tk_boolean:
			return [IDLBooleanItem class];
		case tk_octet:
			return [IDLOctetItem class];
		case tk_char:
		case tk_wchar:
			return [IDLCharItem class];
		case tk_string:
		case tk_wstring:
			return [IDLStringItem class];
	}
}
- defaultItemValue
{
	switch([self kind]){
		default:
			return nil;
		case tk_char:
			return [NSNumber numberWithChar:0];
		case tk_short:
			return [NSNumber numberWithShort:0];
		case tk_ushort:
			return [NSNumber numberWithUnsignedShort:0];
		case tk_long:
			return [NSNumber numberWithLong:0];
		case tk_ulong:
			return [NSNumber numberWithUnsignedLong:0];
		case tk_longlong:
			return [NSNumber numberWithLongLong:0];
		case tk_ulonglong:
			return [NSNumber numberWithUnsignedLongLong:0];
		case tk_float:
			return [NSNumber numberWithFloat:0.0];
		case tk_double:
			return [NSNumber numberWithDouble:0.0];
		case tk_boolean:
			return [NSNumber numberWithBool:NO];
	}
}

- (void)resetItemCell:(NSTextFieldCell*)cell
{
	switch([self kind]){
		case tk_octet:
			[cell setFormatter:[OctetArrayFormatter formatterWithCount:1]];
			[cell setTextColor:[NSColor blackColor]];
			[cell setPlaceholderString:[self name]];
			break;
	}
}

- (BOOL) isItemEditable
{
	switch([self kind]){
		default:
			return YES;
		case tk_boolean:
			return NO;
	}
}
@end

@implementation IDLEnum(IDLItem)

- (Class) itemClass
{
	return [IDLEnumItem class];
}

- (NSCell*) itemDataCell
{
	IDLEnumCell* cell = [[IDLEnumCell alloc]init];
	NSEnumerator* en = [_strings objectEnumerator];
	NSString* key;
	[cell addItemsWithTitles:_strings];
	while(key = [en nextObject]){
		[[cell itemWithTitle:key]setRepresentedObject:[_enumNames objectForKey:key]];
		[[cell itemWithTitle:key]setTag: [[_enumNames objectForKey:key]intValue]];
	}
	return [cell autorelease];
}

@end

@implementation IDLOctetArray(IDLItem)
- (Class) itemClass
{
	return [IDLOctetItem class];
}
- (void)resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setFormatter:[OctetArrayFormatter formatterWithCount:[self arraySize]]];
	[cell setTextColor:[NSColor blackColor]];
	[cell setPlaceholderString:[self name]];
}
- defaultItemValue
{
	NSMutableData* data = [NSMutableData data];
	char b = 0;
	unsigned n = [self arraySize];
	while(n-- > 0)
		[data appendBytes:&b length:1];
	return data;
}

@end

@implementation IDLOctetSequence(IDLItem)
- (Class) itemClass
{
	return [IDLOctetItem class];
}
- (void)resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setFormatter:[OctetSequenceFormatter formatter]];
	[cell setTextColor:[NSColor blackColor]];
	[cell setPlaceholderString:[self name]];
}
- defaultItemValue
{
	return [NSMutableData data];
}

@end

@implementation IDLInterface(IDLItem)
- (Class) itemClass
{
	if([self isAbstract])
		return [IDLAbstractItem class];
	else
		return [IDLIORItem class];
}
- defaultItemValue
{
	return nil;
}
- defaultValueItemValue
{
	return nil;
}

// FIXME - not yet correct
- wrapperValueForValue:value withIDLType:(IDLDefinition*)type
{
	if(![self isAbstract] || value == nil)
		return value;
	else{
		IDLValueObject* obj = [IDLValueObject objectWithIDLType:type];
		[obj setWrappedValue:value];
		return obj;
	}
}

- (IDLItem *) createValueItemWithName:(NSString*)name
{
	if([self isAbstract])
		return nil;
	else
		return [self createIDLItemWithName:name];
}

- (void)resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}

@end

@implementation IDLAny(IDLItem)
- (Class) itemClass
{
	return [IDLAnyItem class];
}
- (BOOL) isLeaf
{
	return NO;
}
- defaultItemValue
{
	return [IDLAnyObject objectWithIDLType:nil];
}
- (void)resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}
@end

@implementation IDLUnion(IDLItem)
- (Class) itemClass
{
	return [IDLUnionItem class];
}
- (BOOL) isLeaf
{
	return NO;
}

- defaultItemValue
{
    IDLSwitchCase* caseObject = _defaultCase;
    NSMutableDictionary* dict = [NSMutableDictionary dictionary];
    if(!caseObject)
        caseObject = [_caseList objectAtIndex:0];
    id switchValue = [caseObject caseValue];
    id caseValue = [[caseObject caseType]defaultItemValue];
    [dict setValue:switchValue forKey:IDLUnionDescriptorKey];
    [dict setValue:caseValue forKey:IDLUnionDataKey];
    return dict;
}

- (void) resetItemCell:(NSTextFieldCell*)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}

@end

@implementation IDLValueType(IDLItem)
- (Class) itemClass
{
	return [IDLValueItem class];
}
- defaultItemValue
{
	return nil;
}
- defaultValueItemValue
{
	return [_support defaultItemValueForIDLType:self];
}

- wrapperValueForValue:value withIDLType:(IDLDefinition *)type
{
	if(_support || value == nil)
		return value;
	else{
		IDLValueObject *obj = [IDLValueObject objectWithIDLType:type];
		[obj setWrappedValue:value];
		return obj;
	}
}

- (NSCell *) itemDataCell
{
	return nil;
}

- (void)resetItemCell:(NSTextFieldCell *)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
	[cell setPlaceholderString:@"<nil>"];
}

- (BOOL) isLeaf
{
	if(_support)
		return [_support isLeaf];
	else
		return YES;
}
- (IDLItem *) createValueItemWithName:(NSString *)name
{
	return [_support createIDLItemWithName:name forIDLType:self];
}

- (BOOL) isItemEditable
{
	return [_support isItemEditable];
}

- (NSArray *) itemArrayForItem:(IDLItem *)item
{
	if([_support respondsToSelector:_cmd])
		return [_support itemArrayForItem:item];
	else
		return nil;
}

- (void) removeValueObserver:(IDLItem*)item forValue:value
{
	if([_support respondsToSelector:_cmd])
		[_support removeValueObserver:item forValue:value];
}

- (void) addValueObserver:(IDLItem*)item forValue:value
{
	if([_support respondsToSelector:_cmd])
		[_support addValueObserver:item forValue:value];
}

- (unsigned) memberCount
{
	return [_support memberCount];
}

@end

@implementation IDLValueClass(IDLItem)

- (IDLItem*) createIDLItemWithName:(NSString*)name forIDLType:(IDLDefinition*)type
{
	IDLItem *item = [[IDLValueClassItem alloc]initWithIDLType:type];
	[item setItemName:name];
	return item;
}

- (void) _appendItemsToArray:(NSMutableArray *)array forItem:(IDLItem *)parent
{
	NSEnumerator *en = [_members objectEnumerator];
	IDLValueMember *member;
	id value = [parent value];
	while(member = [en nextObject]){
		NSString *name = [member name];
		IDLItem *item = [IDLItem itemWithIDLType:[member memberType]];
		[item setItemName:name];
		[item setValue:[value valueForKey:name]];
		[item setEditable:[parent isEditable]];
		[item setController:[parent controller]];
		[array addObject:item];
	}
}

- (void) appendItemsToArray:(NSMutableArray *)array forItem:(IDLItem *)parent
{
	[[_super classSupport]appendItemsToArray:array forItem:parent];
	[self _appendItemsToArray:array forItem:parent];
}
	
- (NSArray *) itemArrayForItem:(IDLItem *)item
{
	NSMutableArray *array = [NSMutableArray array];
	[self appendItemsToArray:array forItem:item];
	return array;
}

- (void) _removeValueObserver:(IDLItem*)item forValue:value
{
	NSEnumerator *en = [_members objectEnumerator];
	IDLValueMember *member;
	while(member = [en nextObject]){
		NSString *name = [member name];
		[value removeObserver:item forKeyPath:name];
	}
}
- (void) _addValueObserver:(IDLItem*)item forValue:value
{
	NSEnumerator *en = [_members objectEnumerator];
	IDLValueMember *member;
	while(member = [en nextObject]){
		NSString *name = [member name];
		[value addObserver:item forKeyPath:name options:0 context:NULL];
	}
}

- (void) removeValueObserver:(IDLItem*)item forValue:value
{
	[[_super classSupport]removeValueObserver:item forValue:value];
	[self _removeValueObserver:item forValue:value];
}

- (void) addValueObserver:(IDLItem*)item forValue:value
{
	[[_super classSupport]addValueObserver:item forValue:value];
	[self _addValueObserver:item forValue:value];
}

- (unsigned) memberCount
{
	return [_super memberCount] + [_members count];
}

- (void) _updateItemValue:(IDLObject*)obj
{
	NSEnumerator *en = [_members objectEnumerator];
	IDLValueMember *member;
	while(member = [en nextObject]){
		IDLDefinition *type = [member memberType];
		NSString *name = [member name];
		[obj setValue:[type defaultItemValue] forKey:name];
	}
	[[_super classSupport]_updateItemValue:obj];
}

- defaultItemValueForIDLType:(IDLDefinition*)type
{
	IDLObject* obj = [IDLObject objectWithIDLType:type];
	[self _updateItemValue:obj];
	return obj;
}

- (BOOL) isLeaf
{
	return NO;
}
- (NSCell *) itemDataCell
{
	return nil;
}
- (void)resetItemCell:(NSTextFieldCell *)cell
{
	[cell setFormatter:nil];
	[cell setTextColor:[IDLItem classColor]];
}
- (BOOL) isItemEditable
{
	return NO;
}
@end

@implementation IDLArrayObject

+ (BOOL)automaticallyNotifiesObserversForKey:(NSString *)key
{
	return NO;
}

- init
{
	return [self initWithArray:nil];
}

- initWithArray:(NSArray*)array
{
	self = [super init];
	_value = array? [[NSMutableArray alloc]initWithArray:array] : [[NSMutableArray alloc]init];
	return self;
}

- initWithCoder:(NSCoder*)coder
{
	self = [super init];
	_value = [[coder decodeObjectForKey:@"value"]retain];
	return self;
}

+ array
{
	return [[[self alloc]init]autorelease];
}

+ arrayWithArray:(NSArray*)array
{
	return [[[self alloc]initWithArray:array]autorelease];
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	[coder encodeObject:_value forKey:@"value"];
}

- (void) dealloc
{
	[_value release];
	[super dealloc];
}

- objectAtIndex:(unsigned) n
{
	id obj = [_value objectAtIndex:n];
	if(obj != [NSNull null])
		return obj;
	else
		return nil;
}

- (NSEnumerator*) objectEnumerator
{
	return [_value objectEnumerator];
}

- (unsigned)count
{
	return [_value count];
}

- (void) makeObjectsPerformSelector:(SEL)sel
{
	[_value makeObjectsPerformSelector:sel];
}

- (void) insertObject:object atIndex:(unsigned)n
{
	if(object == nil)
		object = [NSNull null];
	if(n >= 0 && n <= [_value count]){
		NSIndexSet* set = [NSIndexSet indexSetWithIndex:n];
		[self willChange:NSKeyValueChangeInsertion valuesAtIndexes:set forKey:@"value"];
		[_value insertObject:object atIndex:n];
		[self didChange:NSKeyValueChangeInsertion valuesAtIndexes:set forKey:@"value"];
	}
}

- (void) replaceObjectAtIndex:(unsigned)n withObject:object
{
	if(object == nil)
		object = [NSNull null];
	if(n >= 0 && n <= [_value count]){
		NSIndexSet* set = [NSIndexSet indexSetWithIndex:n];
		[self willChange:NSKeyValueChangeReplacement valuesAtIndexes:set forKey:@"value"];
		[_value replaceObjectAtIndex:n withObject:object];
		[self didChange:NSKeyValueChangeReplacement valuesAtIndexes:set forKey:@"value"];
	}
}

- (void) removeObjectAtIndex:(unsigned)n
{
	if(n >= 0 && n <= [_value count]){
		NSIndexSet* set = [NSIndexSet indexSetWithIndex:n];
		[self willChange:NSKeyValueChangeRemoval valuesAtIndexes:set forKey:@"value"];
		[_value removeObjectAtIndex:n];
		[self didChange:NSKeyValueChangeRemoval valuesAtIndexes:set forKey:@"value"];
	}
}

- (void)removeObjectsAtIndexes:(NSIndexSet *)set
{
	[self willChange:NSKeyValueChangeRemoval valuesAtIndexes:set forKey:@"value"];
	[_value removeObjectsAtIndexes:set];
	[self didChange:NSKeyValueChangeRemoval valuesAtIndexes:set forKey:@"value"];
}

- (void) addObject:object
{
	[self insertObject:object atIndex:[_value count]];
}

@end
