/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDL.h"
#import "CORBAObject_Private.h"
#import "ORB_Private.h"
#import "VXConnection.h"
#import "IDLEncapsulation.h"

@interface NSObject(CORBAObject)
- (void) _setInterface:ifObj;
@end

@implementation IDLInterface(CORBACoding)

- (void) invalidateConnection:(NSNotification*)notif
{
	[self willChangeValueForKey:@"interfaceConnection"];
    [_connection release];
    _connection = nil;
	[self didChangeValueForKey:@"interfaceConnection"];
    [[NSNotificationCenter defaultCenter]removeObserver:self];
}

- interfaceConnection
{
    return _connection;
}

- (void) setInterfaceConnection:aConn
{
    NSEnumerator* en = [_inheritance objectEnumerator];
    IDLInterface* obj;
    [_connection release];
    _connection = [aConn retain];
    while(obj = [en nextObject])
        [obj setInterfaceConnection:aConn];
    if(_connection)
        [[NSNotificationCenter defaultCenter]addObserver:self
            selector:@selector(invalidateConnection:)
            name:ORBInvalidateObject
            object:_connection];
}



- (void) _encodeObject:obj withCoder:(id <CORBACoder>)inv
{
    if(!obj){
        [inv encodeObject:obj];
    }else
    if([obj isKindOfClass:[CORBAObject class]]){
        IDLInterface* ifObj = [obj interface];
        if(!ifObj || [ifObj adoptsInterface:self]) //  interface may be not set
            [inv encodeObject:obj];
        else
            [NSException raise:CORBAMarshalException
                format:@"interface %@ is not derived from %@", [ifObj typeId], [self typeId]];
    }else{
        if([obj respondsToSelector:@selector(interface)]){
            IDLInterface* ifObj = [obj interface];
            if([ifObj isKindOfClass:[IDLInterface class]]){
                if([ifObj adoptsInterface:self]){
                    CORBAObject* proxy = [CORBAObject objectForImplementation:obj withInterface:ifObj];
					VXConnection* currentConnection = [[inv invocationObject]connection];
                    [currentConnection registerImplementationFor:proxy];
                    [inv encodeObject:proxy];
                }else
                    [NSException raise:CORBAMarshalException
                        format:@"interface %@ is not derived from %@", [ifObj typeId], [self typeId]];
            }else
                [NSException raise:CORBAMarshalException
                    format:@"cannot encode object %@ with interface %@", [obj class], [self typeId]];
        }else
            [NSException raise:CORBAMarshalException
                format:@"cannot encode object %@ with interface %@", [obj class], [self typeId]];
    }
}

- (void) encodeObject:obj withCoder:(id <CORBACoder>)inv
{
    if([self isAbstract]){
        BOOL isObjRef = NO;
        if(!obj || [obj isKindOfClass:[CORBAObject class]] ||
                ([obj respondsToSelector:@selector(interface)] &&
                [[obj interface]isKindOfClass:[IDLInterface class]])){
            isObjRef = YES;
            [inv encodeBOOL:isObjRef];
            [self _encodeObject:obj withCoder:inv];
        }else{
            IDLValueType* value = [IDLValueType valuetypeForObject:obj interface:self];
            if(value){
                [inv encodeBOOL:isObjRef];
                // don't want to lookup for valuetype again, call stream
                [inv encodeValue:obj forObjectCoder:value];
            }else
                [NSException raise:CORBAMarshalException 
                    format:@"abstract interface: lookup for valuetype to encode object 0x%x failed", obj];
        }
    }else
    if([self isLocal]){
        [NSException raise:CORBAMarshalException format:@"local interface not supported"];
    }else
        [self _encodeObject:obj withCoder:inv];
    
}

- (void) _decodeObject:(id*)obj withCoder:(id <CORBADecoder>)inv
{
    IDLDefinitionScope* scope = [self rootScope];
    IDLDefinitionScope* current = [IDLDefinitionScope currentScope];
    if(scope != current && scope != [IDLDefinitionScope rootScope])
        [scope set];
    else
    if(current == nil)  // No current scope is set ==> set default
        [[ORB defaultORB]set];
    [inv decodeObject:obj];
    // Object's interface must match 'self' 
    // i.e equal to 'self' or be sub-interface of 'self'
    // We don't check this
    // If obj has no profiles then <nil> object is returned. 
    if([[*obj profiles]count] == 0)
        *obj = nil;
    else
        *obj = [CORBAObject uniqueObjectForObject:*obj];
    // Corba2: Check if local and vended as implementation
    if([*obj isActivated] && [*obj isLocal])
        *obj = [*obj implementation];
}

- (void) decodeObject:(id*)obj withCoder:(id <CORBADecoder>)inv
{
    if([self isAbstract]){
        BOOL isObjRef = NO;
        [inv decodeBOOL:&isObjRef];
        
        if(isObjRef)
            [self _decodeObject:obj withCoder:inv];
        else
            [[IDLValueType valueBase]decodeObject:obj withCoder:inv];
    }else
    if([self isLocal]){
        [NSException raise:CORBAMarshalException format:@"local interface not supported"];
    }else
        [self _decodeObject:obj withCoder:inv];
}

- narrow:anObject
{
    if([anObject respondsToSelector:@selector(_setInterface:)]) // Don't verify typecode
        [anObject _setInterface:self];
    return anObject;
}

- (const char*) objCType
{
    return @encode(id);
}
- (unsigned) argLength
{
    return sizeof(id);
}
- (unsigned) returnValueLength
{
    return sizeof(id);
}

- (void) setValue:value forInvocation:(NSInvocation*)inv atIndex:(int)n
{
    id _val = value;
    [inv setArgument:&_val atIndex:n];
}
- (void) setReturnValue:value forInvocation:(NSInvocation*)inv
{
    id _val = value;
    [inv setReturnValue:&_val];
}
- returnValueFromInvocation:(NSInvocation*)inv
{
    id _val = nil;
    [inv getReturnValue:&_val];
    return _val;
}

- argumentObjectFromInvocation:(NSInvocation*)inv atIndex:(int)n
{
    id _val = nil;
    [inv getArgument:&_val atIndex:n];
    return _val;
}

- argumentObjectOfType:(const char*)argType fromPointer:(void*)ptr
{
    return nil;
}
- (void) setValue:value ofType:(const char*)argType forPointer:(void*)ptr
{
    *((id*)ptr) = value;
}

- (void) _encodeWithCORBACoder:(id <CORBACoder>)coder
{
    [coder encodeString:[self typeId]];
    [coder encodeString:[self name]];
}

- (void) encodeWithCORBACoder:(id <CORBACoder>)coder
{
    long location = [coder locationForDefinition:self];
    if(location != 0 && [IDLDefinition allowsIndirection]) // already encoded -> encode indirection
        [coder encodeIndirection:location];
    else{
        IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
        [coder encodeLong:[self kind]];
        [stream setParentStream:coder location:[coder location]];
        [self _encodeWithCORBACoder:stream];
        [coder encodeData:[stream data]];
    }
}

+ _definitionFromCoder:(IDLEncapsulation*)coder tag:(long)tag
{
    NSString *str;
    NSString* typeId;
    [coder decodeString:&typeId];
    [coder decodeString:&str];
    return [[IDLDefinitionScope currentScope]definitionForTypeId:typeId];
}

+ _definitionFromCoder:(IDLEncapsulation*)coder
{
    return [self _definitionFromCoder:coder tag:tk_objref];
}
// BI_DIR support
- (BOOL) willSendReference
{
    return YES;
}

@end
