/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import <Foundation/Foundation.h>
#import "IDL.h"

@interface IDLParser(Private)
- initRoot;
- (void) addToIdlFiles:(NSString*)file;
- (void) loadInterfaceFromFile:(NSString*)file;
@end

extern int yylex();

@implementation IDLParser

static NSLock* _lock = nil;
static NSString* cppCORBAName = @"__ADORB__";
static IDLParser* _rootParser = nil;

- (void) addToLoadedFiles:(NSString*)path
{
	NSSet* set = [NSSet setWithObject:path];
	[self willChangeValueForKey:@"loadedFiles" withSetMutation:NSKeyValueUnionSetMutation usingObjects:set];
	[loadedFiles unionSet:set];
	[self didChangeValueForKey:@"loadedFiles" withSetMutation:NSKeyValueUnionSetMutation usingObjects:set];
}
- (void) addToIdlFiles:(NSString*)file
{
	[idlFiles addObject:file];
}

+ (NSZone*) zone
{
    return [IDLDefinition zone];
}

+ (void) becomeMultiThreaded:(NSNotification*)notification
{
    _lock = [[NSLock allocWithZone:[self zone]]init];
}

extern int yydebug;

+ (void) initialize
{
    if([IDLParser class] == self){
        NSString* str = [[NSUserDefaults standardUserDefaults]stringForKey:@"debug"];
        if(str && [str rangeOfString:@"p"].length > 0)
            yydebug = 1;

        if([NSThread isMultiThreaded])
            [self becomeMultiThreaded:nil];
        else
            [[NSNotificationCenter defaultCenter] addObserver:self
                selector:@selector(becomeMultiThreaded:)
                name:NSWillBecomeMultiThreadedNotification object:nil];

		_rootParser = [[self allocWithZone:[self zone]]initRoot];
    }
}

- (BOOL) isRoot
{
    return _root;
}

- _initParser
{
    [super init];
    loadedFiles = [[NSMutableSet alloc]init];
    idlFiles = [[NSMutableSet alloc]init];
    _searchPaths = [[NSMutableArray alloc]init];
    return self;
}

- initByParsingIDLFiles:(NSArray*)files searchPaths:(NSArray*)array
{
	[self _initParser];
    rootScope = [[IDLDefinitionScope definitionScope]retain];
    [rootScope defineValue:@"1" forKey:cppCORBAName];
	if(array)
		[_searchPaths addObjectsFromArray:array];
	if([files count] > 0)
		[self loadFiles:files];
	return self;
}

- init
{
	return [self initByParsingIDLFiles:nil searchPaths:nil];
}

- (void) _setupRoot
{
    IDLDefinition* object = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:@"CORBA/Object"];
    [[IDLDefinitionScope rootScope]addDefinition:object withName:@"Object"];
}

- initRoot
{
    NSBundle* bundle = [NSBundle bundleForClass:[self class]];
    NSString* path = [bundle pathForResource:@"default" ofType:@"idl"];
    [self _initParser];
    rootScope = [[IDLDefinitionScope rootScope]retain];
    [rootScope defineValue:@"1" forKey:cppCORBAName];
    if(path){
        NSLock* lock = _lock;
        [lock lock];
        [IDLDefinitionScope _setCurrentPrefix:nil];
        [self loadInterfaceFromFile:path];
        [rootScope applyVersions];
        [rootScope setNamePath:nil];
        [self _setupRoot];
        [lock unlock];
    }
    _root = YES;
    return self;
}

+ rootParser
{
    return _rootParser;
}


- (void) dealloc
{
    [loadedFiles release];
    [rootScope release];
    [idlFiles release];
    [_searchPaths release];
	[_ref release];
    [super dealloc];
}

- (BOOL) isDocumentEdited
{
    return isChanged;
}

- (void) setDocumentEdited:(BOOL)flag
{
    isChanged = flag?YES:NO;
}


- (IDLDefinitionScope*) rootScope
{
    return rootScope;
}

- (void) addSearchPath:(NSString*)path
{
    NSFileManager* fm = [NSFileManager defaultManager];
    BOOL isDir;
    BOOL exists = [fm fileExistsAtPath:path isDirectory:&isDir];
    if(exists && isDir){
        if(![_searchPaths containsObject:path])
            [_searchPaths addObject:path];
    }
}

- (NSString*) locateFile:(NSString*)file directory:(NSString*)dir
{
    NSFileManager* fm = [NSFileManager defaultManager];
    NSString* path= nil;
    BOOL isDir = NO, ok = NO;
    
	// directory 'dir' may be nil
	if(dir){
		path = [dir stringByAppendingPathComponent:file];
		ok = [fm fileExistsAtPath:path isDirectory:&isDir];
    }
    if(!ok){
        NSEnumerator* en = [_searchPaths objectEnumerator];
        while(dir = [en nextObject]){
            path = [dir stringByAppendingPathComponent:file];
            if(ok = [fm fileExistsAtPath:path isDirectory:&isDir])
                break;
        }
    }
    if(ok && !isDir)
        return path;
    else
        NSLog(@"Warning: cannot locate file %@", file);
    return nil;
}

extern IDLDefinitionScope* parseFile(IDLParser* parser, NSString* path);
extern IDLDefinitionScope* parseData(IDLParser* parser, NSData* data);

- (void) parseStream:(NSData*)stream
{
    [IDLDefinitionScope pushScope:rootScope];
    IDLDefinitionScope* scope = parseData(self, stream);
    if(scope != rootScope){
        [[NSNotificationCenter defaultCenter]postNotificationName:@"IDLParseError"
            object:self
            userInfo:[NSDictionary dictionaryWithObject:@"Error while parsing interface\n"
                        forKey:@"message"]];
        while([IDLDefinitionScope popScope]); //clear
    }else
        [IDLDefinitionScope popScope];
    [rootScope awakeFromParse];
}

- (void) parseFile:(NSString*)path
{
    [IDLDefinitionScope pushScope:rootScope];
    IDLDefinitionScope* scope = parseFile(self, path);
    if(scope != rootScope){
        [[NSNotificationCenter defaultCenter]postNotificationName:@"IDLParseError"
            object:self
            userInfo:[NSDictionary dictionaryWithObject:@"Error while parsing interface\n"
                        forKey:@"message"]];
        while([IDLDefinitionScope popScope]); //clear
    }else
        [IDLDefinitionScope popScope];
    [rootScope awakeFromParse];
}


- (void) loadInterfaceFromFile:(NSString*)file
{
    @try{
		[self parseFile:file];
    }@catch(NSException *localException){
		[IDLDefinitionScope clearScopeStack];
		NSLog(@"Exception raised %@: %@\n",
			[localException name],[localException reason]);
    }
}

- (void) loadFiles:(NSArray*)files
{
    NSEnumerator* en = [files objectEnumerator];
    NSString* file;
    NSLock* lock;
    if(_root)
        [NSException raise:@"ParserException" format:@"try to parse files in the root parser"];
    lock = _lock;
    [lock lock];
	@try{
		[IDLDefinitionScope _setCurrentPrefix:nil];
		while(file = [en nextObject]){
			file = [file stringByStandardizingPath];
			NSString* dir = [file stringByDeletingLastPathComponent];
			[_searchPaths addObject:dir]; // Add temporary search path
			[self loadInterfaceFromFile:file];
			[_searchPaths removeObject:dir];
			[self addToIdlFiles:file];
		}
		if([files count] > 0){
			[rootScope applyVersions];
			[rootScope setNamePath:nil];
			[[NSNotificationCenter defaultCenter]postNotificationName:@"ParserDidLoadInterfaces" object:self];
		}
	}@catch(NSException *localException){
		[IDLDefinitionScope clearScopeStack];
		NSLog(@"*** %@: %@", [localException name],[localException reason]);
	}
    [lock unlock];
}

- (NSString*) filename
{
	return _filename;
}

// Save 'idlFiles' array
- (void) saveFilesTo:(NSString*)path
{
	if([idlFiles count] > 0){
		NSDictionary* dict = [NSDictionary dictionaryWithObjectsAndKeys:
			idlFiles, @"files", _searchPaths, @"searchPath", nil];
		[NSKeyedArchiver archiveRootObject:dict toFile:path];
	}
}

// Save as a model
- (void) saveParserTo:(NSString*)path
{
	if([idlFiles count] > 0){
		if([NSKeyedArchiver archiveRootObject:self toFile:path]){
			if(![_filename isEqualToString:path]){
				[_filename release];
				_filename = [path copy];
			}
		}
	}
}

- (NSString*) definedValueForKey:(NSString*)key
{
    return [rootScope definedValueForKey:key];
}

- (void) defineValue:(NSString*)value forKey:(NSString*)key
{
    [rootScope defineValue:value forKey:key];
}

- (void) undefineValueForKey:(NSString*)key
{
    [rootScope undefineValueForKey:key];
}
- (void) undefineKey:(NSString*)key
{
	[self undefineValueForKey:key];
}

- (BOOL) doesDefine:(NSString*)key
{
    return [rootScope definedValueForKey:key] != nil;
}

- (void) setSourceFile:(NSString*)path forObject:object
{
	if(!_ref)
		_ref = [[NSMutableDictionary alloc]init];
	if(path && object)
		[_ref setObject:path forKey:[NSString stringWithFormat:@"%08x", object]];
}
- (NSString*) sourceFileForObject:object
{
	if(object)
		return [_ref objectForKey:[NSString stringWithFormat:@"%08x", object]];
	else
		return nil;
}

@end

@implementation IDLDefinition(Parse)

- (void) awakeFromParse
{
}

@end

@implementation NSDictionary(IDLParser)

- (void) awakeFromParse
{
    NSEnumerator* en = [self objectEnumerator];
    IDLDefinition* obj;
	
    while(obj = [en nextObject]){
        if([obj isKindOfClass:[IDLDefinition class]])
            [obj awakeFromParse];
    }
}

@end

@implementation IDLInterface(Parse)

- (void) awakeFromParse
{
    NSEnumerator* en = [_list objectEnumerator];
    IDLFunction* obj;
	
    while(obj = [en nextObject]){
        if([obj respondsToSelector:@selector(setInterfaceObject:)])
            [obj setInterfaceObject:self];
    }
}

@end

