/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "CORBAException.h"
#import "CORBAObject_Private.h"
#import "IDLException.h"
#import "IDLInterface.h"
#import "IDLDefinitionScope.h"

@implementation CORBAException

NSString* CORBASystemException = @"CORBASystemException";
NSString* CORBAUserException = @"CORBAUserException";
NSString* CORBANotImplemented = @"CORBANotImplemented";
NSString* CORBAForwardRequest = @"CORBAForwardRequestException";

static NSMutableDictionary* _stdList = nil;

+ (void) initialize
{
    if([CORBAException class] == self){
        NSBundle* bundle = [NSBundle bundleForClass:self];
        if(bundle){
            NSString* path = [bundle pathForResource:@"Exceptions" ofType:@"plist"];
			NSDictionary* dict = path? [NSDictionary dictionaryWithContentsOfFile:path] : nil;
            _stdList = dict? [[NSMutableDictionary alloc]initWithDictionary:dict] : [[NSMutableDictionary alloc]init];
        }
    }
}

+ (void) registerStandardExceptionInfo:(NSDictionary*)info forExceptionName:(NSString*)name
{
	[_stdList setObject:info forKey:name];
}

+ (NSDictionary*)standardExceptionInfoForName:(NSString*)cname
{
    return [_stdList objectForKey:cname];
}

+ systemException:(NSString*)cname 
        minorCode:(long)code 
        completed:(CORBACompletionStatus)status
{
    NSString* path = [NSString stringWithFormat:@"CORBA/%@",cname];
    IDLException* excp = (IDLException*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
    if(!excp){
        path = @"CORBA/UNKNOWN";
        excp = (IDLException*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
    }
    if(excp){
        NSNumber* minor = [NSNumber numberWithLong:code];
        NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
            excp,IDLExceptionKey,
            [NSNumber numberWithInt:CORBA_SYSTEM_EXCEPTION], IDLExceptionTypeKey, // system exception
            minor, IDLMinorCodeKey,
            [NSNumber numberWithLong:status], IDLCompletionStatusKey,
            [excp typeId], IDLExceptionIDKey, nil];
        return [self exceptionWithName:CORBASystemException
                        reason:cname userInfo:dict];
    }else
        return [self exceptionWithName:CORBADefinitionException
                    reason:@"Unknown system exception" userInfo:nil];;
}

+ (void) raise:(NSString*)cname minorCode:(long)code completed:(CORBACompletionStatus)status
{
    [[self systemException:cname minorCode:code completed:status]raise];
}

+ userException:(NSString*)namePath info:(NSDictionary*)info
{
    IDLException* excp = (IDLException*)[[IDLDefinitionScope currentScope]definitionForNamePath:namePath];
    if(excp){
        NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
            excp,IDLExceptionKey,
            [NSNumber numberWithInt:CORBA_USER_EXCEPTION], IDLExceptionTypeKey, // User exception
            [excp typeId], IDLExceptionIDKey, nil];
            if(info)
                [dict addEntriesFromDictionary:info];
        return [self exceptionWithName:CORBAUserException
                reason:[excp name] userInfo:dict];
    }else
        return nil;
}

+ exceptionWithName:(NSString*)cname interface:(IDLInterface*)ifObj
{
    IDLException* excp = (IDLException*)[ifObj definitionForNamePath:cname];
    if(excp){
        NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
            excp,IDLExceptionKey,
            [NSNumber numberWithInt:CORBA_USER_EXCEPTION], IDLExceptionTypeKey, // user exception
            [excp typeId],IDLExceptionIDKey, nil];
        return [self exceptionWithName:CORBAUserException reason:cname userInfo:dict];
    }else
        return nil;
}

+ forwardRequestExceptionWithObject:object
{
    IDLException* excp = (IDLException*)[[IDLDefinitionScope rootScope] definitionForNamePath:@"PortableInterceptor/ForwardRequest"];
	NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
		[NSNumber numberWithInt:CORBA_LOCATION_FORWARD], IDLExceptionTypeKey, // LOCATION_FORWARD exception
		object, @"forward",
		excp,IDLExceptionKey,
		[excp typeId],IDLExceptionIDKey, nil];
	return [self exceptionWithName:CORBAForwardRequest reason:@"ForwardRequest" userInfo:dict];
}

+ (BOOL) isForwardRequestException:(NSException*)exception
{
	return [[exception name]isEqualToString:CORBAForwardRequest];
}

- (IDLException*) exceptionDefinition
{
    return [[self userInfo]objectForKey:IDLExceptionKey];
}

+ exceptionWithName:(NSString*)cname object:(CORBAObject*)obj
{
    IDLInterface* ifObj = [obj interface];
    return [self exceptionWithName:cname interface:ifObj];
}

- (void)setValue:(id)value forUndefinedKey:(NSString *)key
{
    IDLException* obj = [self exceptionDefinition];
    if([obj memberWithName:key]){
        [(NSMutableDictionary*)[self userInfo]setObject:value forKey:key];
    }else
        NSLog(@"WARNING: [%@ setValueForKey:] - invalid key value '%@'\n",
              NSStringFromClass([self class]), key);
}

- (id) valueForUndefinedKey:(NSString*)key
{
    return [[self userInfo]valueForKey:key];
}
@end

@implementation CORBAObject(CORBAException)
- (CORBAException*) exceptionWithName:(NSString*)name
{
    return [CORBAException exceptionWithName:name interface:[self interface]];
}

- (CORBAException*) systemException:(NSString*)name minorCode:(long)code completed:(CORBACompletionStatus)status
{
    return [CORBAException systemException:name minorCode:code completed:status];
}

@end

@implementation IDLInterface(CORBAException)

- (CORBAException*) exceptionWithName:(NSString*)name
{
    return [CORBAException exceptionWithName:name interface:self];
}

@end

@implementation NSString(ExcpFIX)

// Standard exceptions listed in 'exceptions.plist' are loaded 
// with 'minor_code_value' as NSString
- (unsigned long) unsignedLongValue
{
    return [self intValue];
}

@end


