/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "VXInvocation.h"
#import "VXConnection.h"
#import "CORBAObject_Private.h"
#import "VXMessage.h"
#import <IDL.h>
#import "IDLServiceContext.h"
#import "IDLTaggedComponent.h"
#import "ORB_Private.h"
#import "IDLPrimitive.h"
#import "IDLTCKind.h"

@interface VXInvocation(RequestHeader)
- (long) request_id;
- (void) _prepareServiceContext;
- (void) _setStringEncodings;
@end

@interface VXInvocation(Invocation) <CORBAInvocation>

@end

@implementation VXInvocation

NSString* CORBATimeoutException = @"CORBATimeoutException";
NSString* CORBACommunicationException = @"CORBACommunicationException";


static BOOL _trace = NO;
static NSLock* _rqlock = nil;
static BOOL _dumpInvocation = NO;
static BOOL _dumpReturn = NO;

+ (void) initialize
{
    if([VXInvocation class] == self){
        NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
        NSString* str = [defaults stringForKey:@"debug"];
        _trace = [defaults boolForKey:@"trace"];
        if(str && [str rangeOfString:@"S"].length > 0)
            _dumpInvocation = YES;
        if(str && [str rangeOfString:@"R"].length > 0)
            _dumpReturn = YES;
        _rqlock = [[NSLock alloc]init];
    }
}

- (void) invalidateConnection:(NSNotification*)notification
{
    [_connection release];
    _connection = nil;
}

static unsigned get_requestId()
{
	static unsigned request = 0;
	[_rqlock lock];
	request = (request + 1)&0x7fffffff;
	if(request == 0)
		request = 1;
	[_rqlock unlock];
	return request;
}

- initWith:(CORBAObject*)obj function:(IDLFunction*)func andArguments:(NSMutableDictionary*)args
{
    self = [super init];
    _function = [func retain];
    _args = [args retain];
    _object = [obj retain];
	_requestContext = [[NSMutableDictionary alloc]init];
	_responseContext = [[NSMutableDictionary alloc]init];
	_lock = [[NSConditionLock alloc]initWithCondition:0];
    _connection = [[_object connection]retain];
	_requestId = get_requestId();
    [_args setValue:[_function name] forKey:@"$operation"];
    if(_connection){
		[self _setStringEncodings];
		[self _prepareServiceContext];
        [[NSNotificationCenter defaultCenter]addObserver:self
					selector:@selector(invalidateConnection:)
					name:VXInvalidateConnection
					object:_connection];
    }else{
        obj = [[_object retain]autorelease]; // obj may be already released
        [self release];
        //[[obj connection]invalidate];
		NSError* error = [obj lastError];
		[[NSException exceptionWithName:CORBACommunicationException
			reason:[NSString stringWithFormat:@"VXInvocation: unable to connect to %@(%hu) for '%@'",
                [obj host], [[obj port]unsignedShortValue], [obj description]]
			userInfo:error? [NSDictionary dictionaryWithObject:error forKey:@"error"] : nil] raise];
    }
    return self;
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
    [_function release];
    [_args release];
    [_object release];
    [_connection release];
	[_response release];
	[_requestContext release];
	[_responseContext release];
	[_lock release];
    [super dealloc];
}

- (void) _handleException:(NSException*)exception
{
	if([CORBAException isForwardRequestException:exception]){
		CORBAObject* forwardObj = [[exception userInfo]objectForKey:@"forward"];
		id delegate = [_object delegate];
		if(![delegate respondsToSelector:@selector(object:willForwardInvocation:toObject:permanently:)] ||
				[delegate object:_object willForwardInvocation:self toObject:forwardObj
							permanently:(_replyStatus == GIOP_LocationForwardPerm)])
		{
			VXInvocation* forwardInv = [forwardObj createInvocationFor:_function withArguments:_args ];
			[forwardInv invoke];
		}
	}else
		[exception raise];
}

- (void) _handleForwardException:(NSException*)exception 
		context:(NSDictionary*)responseContext flowStack:(NSMutableArray*)flowStack
{
	if([flowStack count] > 0){
		NSDictionary* userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
				[_function name], ORBOperationNameKey,
				_object, ORBObjectKey,
				[NSNumber numberWithLong:_requestId], ORBRequestIdKey,
				[NSNumber numberWithUnsignedInt:_replyStatus], ORBReplyStatusKey,
				exception, ORBExceptionKey,
				responseContext, ORBServiceContextKey,
				nil];
		exception = [[ORB notificationCenter]postDidReceiveOther:self userInfo:userInfo flowStack:flowStack];
	}
	[self _handleException:exception];
}

- (void) _handleException:(NSException*)exception
		context:(NSDictionary*)responseContext flowStack:(NSMutableArray*)flowStack
{
	if([flowStack count] > 0){
		NSDictionary* userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
				[_function name], ORBOperationNameKey,
				_object, ORBObjectKey,
				[NSNumber numberWithLong:_requestId], ORBRequestIdKey,
				[NSNumber numberWithUnsignedInt:_replyStatus], ORBReplyStatusKey,
				exception, ORBExceptionKey,
				responseContext, ORBServiceContextKey,
				nil];
		exception = [[ORB notificationCenter]postDidReceiveException:self userInfo:userInfo flowStack:flowStack];
	}
	[self _handleException:exception];
}

- (void) handleException:(NSException*)exception
		context:(NSDictionary*)responseContext flowStack:(NSMutableArray*)flowStack
{
	if([CORBAException isForwardRequestException:exception])
		[self _handleForwardException:exception context:responseContext flowStack:flowStack];
	else
		[self _handleException:exception context:responseContext flowStack:flowStack];
}

- (void) handleMessageException:(VXMessage*)message context:(NSDictionary*)responseContext flowStack:(NSMutableArray*)flowStack
{
    if(_replyStatus == GIOP_LocationForward || _replyStatus == GIOP_LocationForwardPerm){
        id forwardObj = nil;
		[_response alignStreamBody]; // GIOP 1.2 align message buffer
        [[message stream] decodeObject:&forwardObj];

		NSException* exception = [CORBAException forwardRequestExceptionWithObject:forwardObj];
		[self _handleForwardException:exception context:responseContext flowStack:flowStack];
    }else
	if(_replyStatus == GIOP_UserException || _replyStatus == GIOP_SystemException) {
        NSString* typeId;
        NSMutableDictionary* userInfo = nil;
        IDLException* obj;
        NSString* name;
        NSException* exception;
        
		[_response alignStreamBody]; // GIOP 1.2 align message buffer
        [[message stream] decodeString:&typeId];
        obj = [_function exceptionWithTypeID:typeId];
        name = (_replyStatus == GIOP_SystemException) ? 
                    CORBASystemException : CORBAUserException;
        if(obj){
            [obj decodeObject:&userInfo withCoder:[message stream]];
            [userInfo setObject:typeId forKey:IDLExceptionIDKey];
            [userInfo setObject:[NSNumber numberWithInt:_replyStatus] forKey:IDLExceptionTypeKey];
            [userInfo setObject:obj forKey:IDLExceptionKey];
            [userInfo setValue:[obj name] forKey:@"name"];
			NSNumber* status = [userInfo objectForKey:ORBCompletionStatusKey];
			if(status)
				_completionStatus = [status unsignedIntValue];
        }
        exception = [NSException exceptionWithName:name reason:typeId userInfo:userInfo];
		[self _handleException:exception context:responseContext flowStack:flowStack];
    }else
		[NSException raise:CORBANotImplemented format:@"reply status %d is not implemented", _replyStatus];
}

- (void) setResponse:(VXMessage*)message
{
	[_lock lock];
	[_response release];
	_response = [message retain];
	[_response setInvocationObject:self];
	[_lock unlockWithCondition:1];
}

- (void) getResponse
{
	NSTimeInterval t = [self replyTimeout];
	if(t <= 0)
		[_lock lockWhenCondition:1];
	else
		[_lock lockWhenCondition:1  beforeDate:[NSDate dateWithTimeIntervalSinceNow:t]];
	
	[_lock unlock];
	if(!_response){
		NSError* error = [_object lastError];
		NSDictionary* info = error? [NSDictionary dictionaryWithObject:error forKey:@"error"] : nil;
		[[NSException exceptionWithName:CORBATimeoutException
			reason:[NSString stringWithFormat:@"request timed out, interval = %f", t]
			userInfo:info]raise];
	}
}

- (void) invoke
{
    VXMessage* message = [VXMessage requestWithIIOPVersion:[_object iiopVersion]];
    NSString* path = [NSString stringWithFormat:@"GIOP/RequestHeader_%d_%d",
        [message majorVersion],  [message minorVersion]];
    IDLDefinition* def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];

	[message setInvocationObject:self];
	
	NSMutableArray* flowStack = [NSMutableArray array];
	NSDictionary* requestInfo = [NSDictionary dictionaryWithObjectsAndKeys:
		[NSDictionary dictionaryWithDictionary:_requestContext], ORBServiceContextKey,
		[_function name], ORBOperationNameKey,
		_object, ORBObjectKey,
		[NSNumber numberWithLong:_requestId], ORBRequestIdKey,
		[NSNumber numberWithBool:[_function isOneWay]], ORBOnewayKey,
		nil];
	NSException *exception = [[ORB notificationCenter]postWillSendRequest:self userInfo:requestInfo flowStack:flowStack];
	if(exception){
		[self handleException:exception context:nil flowStack:flowStack];
		return;
	}
	
    if(def)
        [[message stream] encodeStruct:self usingObjectCoder:def];
    else
        [NSException raise:CORBADefinitionException format:@"cannot find definition for %@",path];

    if(_trace)
        NSLog(@"SENDING REQUEST <%x> '%@' to '%@:%u' (IIOP%d.%d)\n",
                [self request_id], [_function name],
                [_connection address], [[_connection port]unsignedShortValue],
                [message majorVersion], [message minorVersion]);
    
    [message alignStreamBody];
    [_function encodeObject:_args withCoder:[message stream]];
    if(_dumpInvocation)
        NSLog(@"Sending invocation:\n%@\n\n",[_args description]);
	if([_requestContext objectForKey:[NSNumber numberWithUnsignedInt:ORB_CodeSets]])
		[_connection setCodeSetsSent:YES];
    [_connection sendMessage:message invocation:(![_function isOneWay]?self : nil)];
    
    if(![_function isOneWay]){
		[self getResponse];
		
		NSString* path = [NSString stringWithFormat:@"GIOP/ReplyHeader_%d_%d",
				[_response majorVersion], [_response minorVersion]];
		IDLDefinition* def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
		id replyHeader = nil;

		if(def){
			[def decodeObject:&replyHeader withCoder:[_response stream]];
			_replyStatus = [[replyHeader valueForKey:@"reply_status"]intValue];
		}else
			[NSException raise:CORBADefinitionException format:@"No definition for %@",path];

		NSEnumerator* en = [[replyHeader valueForKey:@"service_context"]objectEnumerator];
		IDLServiceContext* ctxt;
		while(ctxt = [en nextObject])
			[_responseContext setObject:ctxt forKey:[NSNumber numberWithUnsignedInt:[ctxt contextId]]];

		if(_replyStatus != GIOP_NoException)
			[self handleMessageException:_response context:_responseContext flowStack:flowStack];
		else{
			_completionStatus = CORBA_COMPLETION_YES;
			[_response alignStreamBody];
			[_function decodeObject:&_args withCoder:[_response stream]];
			NSDictionary* responseInfo = [NSDictionary dictionaryWithObjectsAndKeys:
				_responseContext, ORBServiceContextKey,
				[_function name], ORBOperationNameKey,
				_object, ORBObjectKey,
				[NSNumber numberWithLong:_requestId], ORBRequestIdKey,
				[NSNumber numberWithLong:_replyStatus], ORBReplyStatusKey,
				[NSNumber numberWithLong:_completionStatus], ORBCompletionStatusKey,
				nil];
			exception = [[ORB notificationCenter]postDidReceiveReply:self userInfo:responseInfo flowStack:flowStack];
			if(exception)
				[self handleException:exception context:_responseContext flowStack:flowStack];
			if(_dumpReturn)
				NSLog(@"Completed invocation:\n%@\n\n", [_args description]);
		}
    }
}

- (NSTimeInterval) replyTimeout
{
    return [_object replyTimeout];
}

- (long) requestId
{
    return _requestId;
}

- (VXConnection*) connection
{
	return _connection;
}

- (NSStringEncoding) stringEncoding
{
	return [_connection stringEncoding];
}

- (NSStringEncoding) wstringEncoding
{
	return [_connection wstringEncoding];
}

@end

@implementation VXInvocation(Invocation)

- (SEL) selector
{
	NSString* string = [_function selectorName];
	if(string)
		return NSSelectorFromString(string);
	else
		return NULL;
}

- (NSString*) selectorString
{
	return [_function selectorName];
}

// Only ObjC type id is implemented
- (void) setObjectReturnValue:(id)value
{
	[_args setObject:value forKey:@"$return"];
}

- returnValue
{
	return [_args objectForKey:@"$return"];
}

- (NSDictionary*) arguments
{
	return _args? [NSDictionary dictionaryWithDictionary:_args] : nil;
}

- (NSDictionary*) requestContext
{
	return [NSDictionary dictionaryWithDictionary:_requestContext];
}

- (IDLServiceContext*) requestContextWithContextId:(unsigned)contextId
{
	return [_requestContext objectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (void) addRequestContext:(IDLServiceContext*)context
{
	[_requestContext setObject:context forKey:[NSNumber numberWithUnsignedInt:[context contextId]]];
}

- (void) removeRequestContextWithContextId:(unsigned)contextId
{
	[_requestContext removeObjectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (NSDictionary*) responseContext
{
	return [NSDictionary dictionaryWithDictionary:_responseContext];
}

- (IDLServiceContext*) responseContextWithContextId:(unsigned)contextId
{
	return [_responseContext objectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (void) addResponseContext:(IDLServiceContext*)context
{
	[_responseContext setObject:context forKey:[NSNumber numberWithUnsignedInt:[context contextId]]];
}

- (void) removeResponseContextWithContextId:(unsigned)contextId
{
	[_responseContext removeObjectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (unsigned) replyStatus
{
	return _replyStatus;
}

- (unsigned) completionStatus
{
	return _completionStatus;
}

- (IDLFunction*) operationObject
{
	return _function;
}

- (CORBAObject*) object
{
	return _object;
}
@end

@implementation VXInvocation(RequestHeader)

- (void) _prepareServiceContext
{
	NSDictionary* context = [IDLServiceContext currentContextDictionary];
	if(context)
		[_requestContext addEntriesFromDictionary:context];
	IDLServiceContext* lpc = [_connection localPointsContext]; // BI_DIR_IIOP
	if(lpc)
		[_requestContext setObject:lpc forKey:[NSNumber numberWithUnsignedInt:[lpc contextId]]];
	
	if(![_connection isCodeSetsSent] && [_connection isStringEncodingsSet]){
		IDLServiceContext* codeSets = [IDLServiceContext codeSetsWithStringEncodings:[_connection stringEncoding]
			:[_connection wstringEncoding]];
		
		if(codeSets)
			[_requestContext setObject:codeSets forKey:[NSNumber numberWithUnsignedInt:[codeSets contextId]]];
    }
}

- (void) _setStringEncodings
{
	if(![_connection isStringEncodingsSet]){
		IDLTaggedComponent* component = [_object taggedComponentWithTag:TAG_CODE_SETS];
		if(component){
			unsigned charCode = [component charEncoding];
			unsigned wcharCode = [component wcharEncoding];
			[_connection setStringEncodings:[IDLServiceContext stringEncodingWithCode:charCode]
				:[IDLServiceContext wstringEncodingWithCode:wcharCode]];
		}else if([IDLServiceContext defaultStringEncoding])
			[_connection setStringEncodings:[IDLServiceContext defaultStringEncoding]:[IDLServiceContext defaultWStringEncoding]];
		//	[_connection setStringEncodings:NSUTF8StringEncoding:NSUnicodeStringEncoding];
	}
}

// Marshaling as RequestHeader

- (NSArray*) service_context
{
	return [_requestContext allValues];
}

- (long) request_id
{
    return [self requestId];
}

- response_expected
{
	BOOL b = ![_function isOneWay];
	return [IDLPrimitive objectForValue:&b ofKind:tk_boolean];
//    return [NSNumber numberWithBool:![_function isOneWay]];
}
- (NSData*)reserved
{
    unsigned char b[] = "\0\0\0";
    return [NSData dataWithBytes:b length:3]; // octet[] mapped to NSData
}
- (NSData*)object_key
{
    return [_object objectKey];
}
- (NSString*)operation
{
    return [_function name];
}
- (NSData*) requesting_principal
{
    return nil;
}
- (NSDictionary*)target // IDL Union
{
    // Always use 'case KeyAddr'. 
    return [NSDictionary dictionaryWithObjectsAndKeys:
        [NSNumber numberWithShort:0], IDLUnionDescriptorKey,
        [_object objectKey], IDLUnionDataKey, nil];
}

- response_flags
{
    char b = [[self response_expected]boolValue]?0x03 : 0x00;
	return [IDLPrimitive objectForValue:&b ofKind:tk_octet]; // octet is treated as unsigned char
//    return [NSData dataWithBytes:&b length:1];
}

@end
